##############################################################################################
#
#  Worker Threads loosely based on QRunnable by Martin Fitzpatrick
#  https://www.pythonguis.com/tutorials/multithreading-pyside6-applications-qthreadpool/
#  License: MIT
#
##############################################################################################

import sys
import traceback
from dataclasses import dataclass
from typing import Callable

from PySide6.QtCore import QRunnable, Slot, Signal, QObject


@dataclass(frozen=True, slots=True)
class WorkerError:
    exception_type: type
    value: Exception
    traceback: str
    args: tuple | None = None
    kwargs: dict | None = None

    def __str__(self) -> str:
        return f"{self.exception_type}: {self.traceback}\n{self.value}"


class WorkerSignals(QObject):
    """
    Defines the signals available from a running worker thread.

    Supported signals are:

    finished
        The initial parameters passed to the worker when it was started, args and kwargs.

    error
        an instance of WorkerError.

    result
        object data returned from processing, if anything.

    progress
        anything to indicate the current state.

    aborted
        The initial parameters passed to the worker when it was started, args and kwargs.

    """

    finished = Signal(tuple)
    error = Signal(WorkerError)
    result = Signal(object)
    progress = Signal(object)
    aborted = Signal(tuple)


class SharableFlag:
    def __init__(self, initial_value: bool = False) -> None:
        self._flag = initial_value

    def get(self) -> bool:
        return self._flag

    def set(self, value: bool) -> None:
        self._flag = value


class Abort(Exception):
    """
    Exception to abort the worker.
    """

    pass


class Worker(QRunnable):
    """
    Worker thread

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    DO NO. I REPEAT. DO NOT PUT @Slot() DECORATORS ON THE FUNCTIONS THAT YOU
    CONNECT TO THE SIGNALS. YOU WILL ONLY BE REWARDED WITH SEGFAULTS.

    :param callback: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    :param args: Arguments to pass to the callback function.
    :param no_progress_callback: [Optional] If True, the progress_callback will not be added to the kwargs.
    :param abort_signal: [Optional] A signal that will be emitted when the thread should abort.
    :param kwargs: Keywords to pass to the callback function.
    """

    aborted: SharableFlag

    def __init__(
        self,
        fn: Callable,
        *args,
        no_progress_callback: bool = False,
        abort_signal: Signal | None = None,
        **kwargs,
    ):
        QRunnable.__init__(self)

        # Store constructor arguments (re-used for processing).
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

        # If the abort signal is received, the abort flag is set to true.
        # The worker process must abort itself when the flag is true.
        self.aborted = SharableFlag(False)

        # Add the callback to our kwargs.
        if not no_progress_callback:
            self.kwargs["progress_callback"] = self.signals.progress

        # If an abort signal is provided, provide the sharable flag to the worker.
        if abort_signal is not None:
            self.kwargs["abort_flag"] = self.aborted
            abort_signal.connect(self.abort)

    @Slot()
    def run(self) -> None:
        """
        Initialise the runner function with passed args, kwargs.
        """

        # Retrieve args/kwargs here; and fire processing using them.
        # The function should expect the keyword arguments:
        # progress_callback and abort_flag unless disabled in the constructor.

        # After running, we need to check if the signals still exist. They may have been deleted in
        # the instance that the program was closed while the worker was running.
        # Otherwise the Python runtime crashes, which is ok in this state, but it just isn't very nice.
        try:
            try:
                result = self.fn(*self.args, **self.kwargs)
            except Abort:
                self.signals.aborted.emit((self.args, self.kwargs))
            except:
                # traceback.print_exc()  # Disabled because we handle showing the error in whatever
                # called the worker. This also prevents the runtime errors from getting printed, which
                # are caught in an outer try block.
                exception_type, value = sys.exc_info()[:2]
                self.signals.error.emit(
                    WorkerError(
                        exception_type, value, traceback.format_exc(), self.args, self.kwargs
                    )
                )
            else:
                self.signals.result.emit(result)  # Return the result of the processing.
            finally:
                self.signals.finished.emit((self.args, self.kwargs))  # Done
        except RuntimeError:
            # The signals were deleted, so we can't emit them.
            # This is fine, we just don't need to emit anything.
            pass

    @Slot()
    def abort(self) -> None:
        self.aborted.set(True)



import PySide6.QtWidgets as Qw
import PySide6.QtGui as Qg
import PySide6.QtCore as Qc
from PySide6.QtCore import Qt


class CTooltipLabel(Qw.QLabel):
    """
    A label that displays a help-hint icon and shows a tooltip when clicked.
    This is to make little helper icons more consistent.
    """

    icon_name: str

    def __init__(self, parent=None, tooltip: str = "", icon_name: str = "help-hint") -> None:
        super(CTooltipLabel, self).__init__(parent)

        # Display the help-hint icon and no text.
        self.icon_name = icon_name
        self.load_icon()
        self.setText("")
        if tooltip:
            self.setToolTip(tooltip)

        # Make the label focusable to receive keyboard events
        self.setFocusPolicy(Qt.StrongFocus)

    def load_icon(self) -> None:
        """
        Load the display icon.
        """
        self.setPixmap(Qg.QIcon.fromTheme(self.icon_name).pixmap(16, 16))

    def setText(self, text: str) -> None:
        """
        Ignore all attempts to set the text.
        """
        pass

    def keyPressEvent(self, event) -> None:
        """
        Show tooltip when the Enter or Space key is pressed.
        """
        if event.key() in (Qt.Key_Return, Qt.Key_Space):
            # Showing tooltip at the center of the label when triggered by keyboard.
            self.showTooltipAtPosition(self.rect().center())

    def showTooltipAtPosition(self, pos: Qc.QPoint) -> None:
        """
        Show the tooltip at the given local position.
        """
        global_pos = self.mapToGlobal(pos)
        Qw.QToolTip.showText(global_pos, self.toolTip())



"""
This file contains a slightly modified version of the original function,
and serves as the only link to the comic_text_detector project:
https://github.com/dmMaze/comic-text-detector
(License: GNU General Public License v3.0)

Should a newer version of the comic_text_detector project be released,
replacing the files in the pcleaner/comic_text_detector folder with the
newer version should be sufficient to update the code, provided that the
coupling between this function and the rest of the code is not broken.

The code present in the pcleaner/comic_text_detector folder is a
slimmed-down version of the original project, containing only the
necessary files to run the model and the function below.
"""

import json
import multiprocessing as mp
from math import floor, ceil
from pathlib import Path
from uuid import uuid4

import cv2
import numpy as np
import torch
from loguru import logger
from tqdm import tqdm

import pcleaner.config as cfg
from pcleaner.comic_text_detector.inference import TextDetector
from pcleaner.comic_text_detector.utils.io_utils import imwrite, NumpyEncoder
from pcleaner.comic_text_detector.utils.textmask import REFINEMASK_ANNOTATION


def model2annotations(
    config_general: cfg.GeneralConfig,
    config_detector: cfg.TextDetectorConfig,
    model_path: Path,
    img_list: list[Path],
    save_dir: Path,
):
    """
    Run the model on a directory of images and produce the following
    for each image inside the save_dir directory:
    - A copy of the original image as a .png file.
    - A .png file containing the text mask, filename: <image_name>_mask.png.
    - A .json file containing each box of text, as well as other metadata,
      filename: <image_name>.json.

    For this modified version, include the image name and mask name in the
    json file.

    :param config_general: General configuration, part of the profile.
    :param config_detector: Text detector configuration, part of the profile.
    :param model_path: Path to the model file. This ends either in .pt or .onnx (torch or cv2 format).
    :param img_list: Path or a list of paths to an image or directory of images.
    :param save_dir: Path to the directory where the results will be saved.
    :return:
    """

    device = "cuda" if model_path.suffix == ".pt" else "cpu"
    print(f"Using device for text detection model: {device}")
    # Determine the number of processes to use
    num_processes = min(config_detector.concurrent_models, len(img_list))
    print(f"Using {num_processes} processes for text detection.")

    if num_processes > 1:
        mp.freeze_support()
        mp.set_start_method("spawn")

        with mp.Pool(num_processes) as pool:
            # Distribute the images evenly among the processes.
            batches = [list() for _ in range(num_processes)]
            for i, img_path in enumerate(img_list):
                batches[i % num_processes].append(img_path)

            args = [
                (
                    batch,
                    model_path,
                    device,
                    save_dir,
                    config_general.input_height_lower_target,
                    config_general.input_height_upper_target,
                )
                for batch in batches
            ]

            for _ in tqdm(pool.imap_unordered(process_image_batch, args), total=len(args)):
                pass  # do nothing, just iterate through the results

    else:
        model = TextDetector(model_path=str(model_path), input_size=1024, device=device)

        for index, img_path in enumerate(tqdm(img_list)):
            process_image(
                img_path,
                model,
                save_dir,
                config_general.input_height_lower_target,
                config_general.input_height_upper_target,
            )


def process_image_batch(args) -> None:
    img_batch, model_path, device, save_dir, height_target_lower, height_target_upper = args
    model = TextDetector(model_path=str(model_path), input_size=1024, device=device)
    for img_path in img_batch:
        process_image(img_path, model, save_dir, height_target_lower, height_target_upper)

    del model

    if device == "cuda":
        # Release CUDA resources.
        torch.cuda.ipc_collect()
        torch.cuda.empty_cache()


def process_image(
    img_path: Path,
    model: TextDetector,
    save_dir: Path,
    height_target_lower: int,
    height_target_upper: int,
    skip_text_detection: bool = False,
    uuid: str = None,
):
    """
    Process a single image using the TextDetector model.
    This generates a mask and a json file containing the text boxes.
    These both belong in the cache directory.

    :param img_path: The path to the image to process.
    :param model: The TextDetector model.
    :param save_dir: The directory where the results will be saved.
    :param height_target_lower: The lower target height of the input image.
    :param height_target_upper: The upper target height of the input image.
    :param skip_text_detection: If True, skip the text detection step and only
        save the scaled input image as a png.
    :param uuid: The uuid to use for the image. If None, a new uuid will be generated.
    """

    img = read_image(img_path)
    img, image_scale = resize_to_target(img, height_target_lower, height_target_upper)

    # Prepend an index to prevent name clobbering between different files.
    if uuid is None:
        prefix = f"{uuid4()}_"
    else:
        prefix = f"{uuid}_"

    img_name = prefix + img_path.stem
    maskname = img_name + "_mask.png"

    # Make names absolute paths.
    img_name = str((save_dir / (img_name + ".png")).absolute())
    maskname = str((save_dir / maskname).absolute())

    # Save a scaled copy of the image.
    imwrite(img_name, img)
    if skip_text_detection:
        # If we're only saving the scaled image, we're done.
        return

    mask, mask_refined, blk_list = model(
        img, refine_mode=REFINEMASK_ANNOTATION, keep_undetected_mask=True
    )
    blk_xyxy = []
    blk_dict_list = []
    for blk in blk_list:
        blk_xyxy.append(blk.xyxy)
        blk_dict_list.append(blk.to_dict())

    # Inject the img_name.png and mask name and original path into the json.
    data = {
        "image_path": img_name,
        "mask_path": maskname,
        "original_path": str(img_path),
        "scale": image_scale,
        "blk_list": blk_dict_list,
    }
    # Remove the suffix and add _raw.json
    json_path = Path(img_name).with_suffix(".json")
    json_path = json_path.with_stem(json_path.stem + "#raw")
    logger.debug(f"Saving json file to {json_path}")
    with open(json_path, "w", encoding="utf8") as f:
        json.dump(data, f, ensure_ascii=False, cls=NumpyEncoder, indent=4)
    imwrite(maskname, mask_refined)


def read_image(path: Path | str) -> np.ndarray:
    """
    Read image from path, scaling it if necessary.
    Then return an array of the image.

    :param path: Image path
    :return: Image array
    """

    return cv2.imdecode(np.fromfile(str(path), dtype=np.uint8), cv2.IMREAD_COLOR)


def calculate_new_size_and_scale(
    width: int, height: int, height_target_lower: int, height_target_upper: int
) -> tuple[int, int, float]:
    """
    Calculate the new dimensions and scale factor for resizing an image.
    Prefer the largest whole integer scale factor to scale down by.
    Never scales up and skips scaling if one of the dimensions is <= 0.

    :param width: The current width of the image.
    :param height: The current height of the image.
    :param height_target_lower: The lower target height of the input image.
    :param height_target_upper: The upper target height of the input image.
    :return: Tuple containing new width, new height, and scale factor.
    """
    if height_target_lower <= 0 or height_target_upper <= 0 or height <= height_target_upper:
        return width, height, 1.0

    # Check if both targets are the same, or the lower one is greater, requiring an exact size.
    if height_target_lower >= height_target_upper:
        scale = height_target_upper / height
        new_width = round(width * scale)
        new_height = height_target_lower
    else:
        # Otherwise, check bounds and choose the largest whole integer scale factor, if possible.
        # Using the inverse scale factor to determine denominators like 1/2, 1/3, etc.
        inv_scale_lower = height / height_target_lower
        inv_scale_upper = height / height_target_upper
        inv_scale_upper_nearest = ceil(inv_scale_upper)
        # Check if the nearest upper bound is within the range.
        if inv_scale_upper_nearest <= inv_scale_lower:
            scale = 1 / inv_scale_upper_nearest
            new_width = round(width * scale)
            new_height = round(height * scale)
        else:
            # Otherwise just try to choose the largest size that results in a height that is a multiple of 4.
            max_height = round(height / inv_scale_upper)
            min_height = round(height / inv_scale_lower)
            # Find the largest multiple of 4 that is within the range.
            new_height = floor(max_height / 4) * 4
            if new_height < min_height:
                # If the largest multiple of 4 is still too small, just use the upper bound.
                new_height = max_height
            scale = new_height / height
            new_width = round(width * scale)

    return new_width, new_height, scale


def resize_to_target(
    image: np.ndarray, height_target_lower: int, height_target_upper: int
) -> tuple[np.ndarray, float]:
    """
    Resize the image to fall within the target height range.
    Prefer the largest whole integer scale factor to scale down by.
    Never scales up and skips scaling if one of the dimensions is <= 0.

    :param image: The image to resize as a numpy array, loaded by cv2.
    :param height_target_lower: The lower target height of the input image.
    :param height_target_upper: The upper target height of the input image.
    :return: The resized image and the scale factor.
    """
    height, width, _ = image.shape  # As a matrix, it returns the height (columns) first.
    new_width, new_height, scale = calculate_new_size_and_scale(
        width, height, height_target_lower, height_target_upper
    )

    return cv2.resize(image, (new_width, new_height), interpolation=cv2.INTER_AREA), scale



import PySide6.QtWidgets as Qw

import pcleaner.config as cfg
import pcleaner.model_downloader as md
from pcleaner.gui.ui_generated_files.ui_SetupGreeter import Ui_SetupGreeter


class SetupGreeter(Qw.QDialog, Ui_SetupGreeter):
    """
    A greeting dialog explaining that Panel Cleaner needs to download additional files before
    it can start. The user can then either proceed or abort.
    """

    def __init__(
        self,
        parent: Qw.QWidget | None = None,
        config: cfg.Config | None = None,
    ):
        """
        Initialize the widget.

        :param parent: The parent widget.
        """
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        # Update the placeholders.
        self.label_text_detector_path.setText(str(config.get_model_cache_dir()))
        self.label_ocr_path.setText(str(md.get_ocr_model_directory()))



import multiprocessing as mp
from functools import partial
from pathlib import Path
from typing import Sequence

import torch

import pcleaner.config as cfg
import pcleaner.gui.image_file as imf
import pcleaner.gui.worker_thread as wt
from pcleaner.comic_text_detector.inference import TextDetector
from pcleaner.ctd_interface import process_image


# It's simply slower for anything less than this many images.
MIN_IMAGES_FOR_MULTIPROCESSING = 20


def model2annotations_gui(
    config_general: cfg.GeneralConfig,
    config_detector: cfg.TextDetectorConfig,
    text_detector_model_path: Path,
    img_list: Sequence[imf.ImageFile],
    save_dir: Path,
    no_text_detection: bool,
    partial_progress_data: partial,
    progress_callback: imf.ProgressSignal | None,
    abort_flag: wt.SharableFlag,
):
    """
    Run the model on a directory of images and produce the following
    for each image inside the save_dir directory:
    - A copy of the original image as a .png file.
    - A .png file containing the text mask, filename: <image_name>_mask.png.
    - A .json file containing each box of text, as well as other metadata,
      filename: <image_name>.json.

    For this modified version, include the image name and mask name in the
    json file.

    The partial progress data must have the following arguments pre-filled:
    - The total number of images to process.
    - The target output to reach.
    - The current step.

    Note on Aborting:
    The sharable flag is passed to this function because it is what iterates over the images.
    It must not interrupt the process for a single image, but should prevent the next image
    from being processed.
    All this function needs to do is raise the Abort exception, the calling function will send
    the appropriate progress update to the rest of the gui.

    :param config_general: General configuration, part of the profile.
    :param config_detector: Text detector configuration, part of the profile.
    :param text_detector_model_path: Path to the model file. This ends either in .pt or .onnx (torch or cv2 format).
    :param img_list: Path or a list of paths to an image or directory of images.
    :param save_dir: Path to the directory where the results will be saved.
    :param no_text_detection: If True, the text detection step is skipped. This is to only get the "input" output.
    :param partial_progress_data: A partial function to create a ProgressData object.
        You must only add the current iteration to the partial function to construct a valid
        ProgressData object.
    :param progress_callback: A callback to report progress to the gui.
    :param abort_flag: A flag to indicate if the process should abort.
    """

    if progress_callback is not None:
        progress_callback.emit(partial_progress_data(imf.ProgressType.begin_step))

    def inc_progress() -> None:
        if progress_callback is not None:
            progress_callback.emit(partial_progress_data(imf.ProgressType.incremental))

    num_processes = min(config_detector.concurrent_models, len(img_list))

    if (
        num_processes > 1
        and not no_text_detection
        and len(img_list) > MIN_IMAGES_FOR_MULTIPROCESSING
    ):
        mp.freeze_support()
        mp.set_start_method("spawn")

        with mp.Pool(num_processes) as pool:
            # Distribute the images evenly among the processes.
            # The batches is a list of lists of tuples, where each tuple contains the image path and uuid.
            batches: list[list[tuple[Path, str]]] = [list() for _ in range(num_processes)]
            for i, img_obj in enumerate(img_list):
                batches[i % num_processes].append((img_obj.path, img_obj.uuid))

            device = "cuda" if text_detector_model_path.suffix == ".pt" else "cpu"
            args = [
                (
                    batch,
                    text_detector_model_path,
                    device,
                    save_dir,
                    config_general.input_height_lower_target,
                    config_general.input_height_upper_target,
                    inc_progress,
                )
                for batch in batches
            ]

            for _ in pool.imap_unordered(process_image_batch, args):
                if abort_flag.get():
                    raise wt.Abort()

    else:
        # Load the model.
        cuda = torch.cuda.is_available()
        device = "cuda" if cuda else "cpu"
        model = TextDetector(
            model_path=str(text_detector_model_path), input_size=1024, device=device
        )

        for img_obj in img_list:
            if abort_flag.get():
                raise wt.Abort()

            process_image(
                img_obj.path,
                model,
                save_dir,
                config_general.input_height_lower_target,
                config_general.input_height_upper_target,
                no_text_detection,
                img_obj.uuid,
            )
            inc_progress()


def process_image_batch(args) -> None:
    img_batch, model_path, device, save_dir, lower_target, upper_target, inc_progress = args
    model = TextDetector(model_path=str(model_path), input_size=1024, device=device)
    for img_path, img_uuid in img_batch:
        process_image(img_path, model, save_dir, lower_target, upper_target, uuid=img_uuid)
        inc_progress()

    del model

    if device == "cuda":
        # Release CUDA resources.
        torch.cuda.ipc_collect()
        torch.cuda.empty_cache()



import difflib
import platform
import subprocess
from itertools import takewhile, groupby
from pathlib import Path

import tifffile
from loguru import logger


def f_plural(value, singular: str, plural: str) -> str:
    """
    Selects which form to use based on the value.
    """
    return singular if value == 1 else plural


def open_file(path: Path) -> None:
    """
    Open any given file with the default application.
    """
    logger.info(f"Opening file {path}")
    try:
        if platform.system() == "Linux":
            subprocess.run(["xdg-open", path])
        elif platform.system() == "Windows":
            subprocess.run(["notepad", path])
        elif platform.system() == "Darwin":
            subprocess.run(["open", path])
    except Exception as e:
        logger.exception(e)


def closest_match(word: str, choices: list[str]) -> str | None:
    """
    Return the closest match for the given word in the list of choices.
    If no good match is found, return None.
    """
    if word in choices:
        return word
    else:
        # Find the closest match using difflib:
        closest = difflib.get_close_matches(word, choices, 1, 0.5)  # 0.6 is the default threshold
        if closest:
            return str(closest[0])
        else:
            return None


def discover_all_images(
    img_paths: str | Path | list[str | Path], supported_extensions: list[str]
) -> tuple[list[Path], list[Path]]:
    """
    Discover all images in the given paths.
    Perform a shallow search in directories, not recursing into subdirectories.

    :param img_paths: A path to a single image, directory, or multiple of either.
    :param supported_extensions: A list of supported image extensions. e.g. [".png", ".jpg"]
    :return: A list of all accepted images found by path only, and a list of rejected tiff files,
        if they have an unsupported format.
    """
    img_list: list[Path] = []

    # Wrap single paths in a list.
    if isinstance(img_paths, str):
        img_paths = [Path(img_paths)]
    elif isinstance(img_paths, Path):
        img_paths = [img_paths]
    elif isinstance(img_paths, list):
        # Convert all strings to paths.
        img_paths = [Path(path) for path in img_paths]
    else:
        raise TypeError(f"Invalid type for img_paths: {type(img_paths)}")

    for img_path in img_paths:
        if img_path.is_dir():
            img_list.extend(find_all_images_shallow(img_path, supported_extensions))
        elif img_path.is_file() and img_path.suffix.lower() in supported_extensions:
            img_list.append(img_path)
        elif img_path.is_file():
            raise ValueError(f"Unsupported image format: {img_path.suffix} for {img_path}")
        else:
            raise FileNotFoundError(f"Image path {img_path} does not exist.")

    # Ensure all paths are absolute.
    img_list = [path.resolve() for path in img_list]

    # Filter out 5 channel TIFFs, as they are not supported.
    rejected_tiffs = [path for path in img_list if is_5_channel_tiff(path)]
    img_list = [path for path in img_list if not is_5_channel_tiff(path)]

    return img_list, rejected_tiffs


def find_all_images_shallow(img_dir: Path, supported_extensions: list[str]) -> list[Path]:
    """
    Non-recursive search for all images in the given directory.

    :param img_dir: The directory to search.
    :param supported_extensions: A list of supported image extensions. e.g. [".png", ".jpg"]
    :return: A list of all images found in the given directory.
    """
    image_list: list[Path] = []
    for file_path in img_dir.glob("*"):
        file_suffix = file_path.suffix
        if file_suffix.lower() not in supported_extensions:
            continue
        else:
            image_list.append(file_path)
    return image_list


def is_5_channel_tiff(path: Path) -> bool:
    """
    Returns True if the given file is a TIFF image with a 5.1 channel, False otherwise.
    """
    # Check suffix first.
    if path.suffix not in [".tif", ".tiff"]:
        return False
    # Try opening the file as a TIFF image
    with tifffile.TiffFile(path) as tif:
        logger.debug("Tiff data:\n" + str(tif.pages[0].tags))
        # Check if the TIFF image has 5 channels.
        try:
            if tif.pages[0].tags["SamplesPerPixel"].value == 5:
                logger.warning(
                    f"Found a 5 channel TIFF image: {path}. These are not supported, the image will be skipped."
                )
                return True
        except KeyError:
            pass

    return False


def all_equal(iterable) -> bool:
    """
    Checks if all items in a list are the same using itertools.groupby.
    It works because groupby will produce a single group if and only if all items are the same.
    This groupby function is highly performant.

    :param iterable: An iterable.
    :return: True if all items are the same, False otherwise.
    """
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def trim_prefix_from_paths(paths: list[Path]) -> list[Path]:
    """
    For a given list of Path objects, shorten them by removing the longest common prefix.
    This will only split along directory boundaries, so the paths will remain valid, but no
    longer relative to the original directory.
    It will not trim the actual file name.

    :param paths: A list of paths to shorten.
    :return: A list of shortened paths as Path objects.
    """
    if not paths:
        return []

    # Split all Path objects into their constituent parts.
    parts = [list(p.parts[:-1]) for p in paths]

    # Calculate how many of the first parts are the same for all paths.
    prefix_len = sum(1 for _ in takewhile(all_equal, zip(*parts)))

    # Shorten paths by removing the common prefix
    short_paths = [Path(*p.parts[prefix_len:]) for p in paths]

    return short_paths


def common_path_parent(paths: list[Path]) -> Path:
    """
    For a given list of Path objects, shorten them by removing the longest common prefix.
    This will only split along directory boundaries, so the paths will remain valid, but no
    longer relative to the original directory.
    It will not trim the actual file name.

    :param paths: A list of paths to shorten.
    :return: A list of shortened paths as Path objects.
    """
    if not paths:
        return Path()

    # Special case: if any path is root (its parent is itself), return root
    root_path = next((p for p in paths if p.parent == p), None)
    if root_path is not None:
        return root_path

    # Split all Path objects into their constituent parts.
    parts = [list(p.parts[:-1]) for p in paths]

    # Calculate how many of the first parts are the same for all paths.
    prefix_len = sum(1 for _ in takewhile(all_equal, zip(*parts)))

    prefix = Path(*paths[0].parts[:prefix_len])

    return prefix



import os
import re
import time
from pathlib import Path

import PySide6.QtCore as Qc
import PySide6.QtWidgets as Qw
import humanfriendly as hf
import requests
import torch
from PySide6.QtCore import Signal, QObject, Slot
from attr import frozen
from humanfriendly import format_timespan, parse_size
from loguru import logger
from manga_ocr import MangaOcr

import pcleaner.config as cfg
import pcleaner.gui.gui_utils as gu
import pcleaner.gui.worker_thread as wt
import pcleaner.model_downloader as md
from pcleaner.gui.ui_generated_files.ui_ModelDownloader import Ui_ModelDownloader


PROGRESS_UPDATE_INTERVAL = 0.5


def format_size(size: int | float) -> str:
    """
    Format a size in bytes to a human readable string.

    :param size: The size in bytes.
    :return: The size as a human readable string.
    """
    return hf.format_size(size, binary=True)


@frozen
class ProgressData:
    """
    A data class for holding the progress data of a download.

    - percentage: The percentage of the download, from 0 to 100.
    - downloaded_size: The number of bytes downloaded so far.
    - file_size: The total size of the file in bytes.
    - speed: The download speed in bytes per second.
    - eta: The estimated time until the download is finished, in seconds.
        When no accurate estimate can be given yet, it will be -1.
    - file_name: [Optional] The name of the file being downloaded. Only used for the ocr model.
    """

    percentage: int
    downloaded_size: int
    file_size: int
    speed: float
    eta: int
    file_name: str | None = None


class DownloadSignals(QObject):
    progress_signal = Signal(ProgressData)


class ModelDownloader(Qw.QDialog, Ui_ModelDownloader):
    """
    A widget showing the current download progress of the model.
    """

    model_name: str = ""
    model_path: Path | None = None

    text_detector_downloaded: bool = False
    ocr_downloaded: bool = False
    encountered_error: bool = False

    def __init__(
        self,
        parent: Qw.QWidget | None = None,
        config: cfg.Config = None,
        need_text_detector: bool = True,
        need_ocr: bool = True,
    ):
        """
        Initialize the widget.

        :param parent: The parent widget.
        :param config: The config to use.
        :param need_text_detector: If True, download the text detector model.
        :param need_ocr: If True, download the ocr model.
        """
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.text_detector_downloaded = not need_text_detector
        self.ocr_downloaded = not need_ocr

        self.buffer = []

        if need_text_detector:
            self.download_text_detector(config)
        else:
            temp_progress_data = ProgressData(100, 0, 0, 0, -1)
            self.model_name = "Already downloaded"
            self.show_text_detection_progress(temp_progress_data)

        if need_ocr:
            self.download_ocr()
        else:
            temp_progress_data = ProgressData(100, 0, 0, 0, -1, "Already downloaded")
            self.show_ocr_progress(temp_progress_data)

    def check_finished(self) -> None:
        if self.text_detector_downloaded and self.ocr_downloaded:
            if not self.encountered_error:
                logger.info("Finished downloading all models.")
                self.accept()
            else:
                logger.error("Failed to download all models.")
                self.reject()

    # ============================= Text Detector =============================

    def download_text_detector(self, config: cfg.Config) -> None:
        # Check if we need the cuda or cpu model.
        if torch.cuda.is_available():
            model_url = md.MODEL_URL + md.TORCH_MODEL_NAME
            sha_hash = md.TORCH_SHA256
            self.model_name = "Text Detector model (CUDA)"
        else:
            model_url = md.MODEL_URL + md.CV2_MODEL_NAME
            sha_hash = md.CV2_SHA256
            self.model_name = "Text Detector model (CPU)"

        # Init the ui.
        temp_progress_data = ProgressData(0, 0, 0, 0, -1)
        self.show_text_detection_progress(temp_progress_data)

        # Start the download in a separate thread.
        threadpool = Qc.QThreadPool.globalInstance()
        signals = DownloadSignals()
        signals.progress_signal.connect(self.show_text_detection_progress)

        worker = wt.Worker(
            download_file,
            model_url,
            config.get_model_cache_dir(),
            signals,
            sha_hash,
            no_progress_callback=True,
        )
        worker.signals.finished.connect(self.text_detection_finished)
        worker.signals.result.connect(self.text_detection_result)
        worker.signals.error.connect(self.text_detection_error)
        threadpool.start(worker)

    @Slot(ProgressData)
    def show_text_detection_progress(self, progress_data: ProgressData) -> None:
        """
        Update the progress bar.

        :param progress_data: The progress data.
        """
        self.label_model_name.setText(self.model_name)
        self.label_model_size.setText(
            f"{format_size(progress_data.downloaded_size)} / {format_size(progress_data.file_size)}"
        )
        self.label_model_speed.setText(
            f"{format_size(progress_data.speed)}/s, ETA: {format_timespan(progress_data.eta, ) if progress_data.eta >= 0 else '—'}"
        )
        self.progressBar_model.setValue(progress_data.percentage)

    @Slot(wt.WorkerError)
    def text_detection_error(self, worker_error: wt.WorkerError) -> None:
        """
        Show an error message.

        :param worker_error: The error message.
        """
        self.label_model_speed.setText("—")
        self.label_model_size.setText("")
        self.encountered_error = True
        logger.exception(worker_error.value)
        gu.show_warning(self, "Download Failed", str(worker_error.value))

    @Slot(Path)
    def text_detection_result(self, model_path: Path | None) -> None:
        """
        Save the path to the downloaded model.

        :param model_path: The path to the downloaded model.
        """
        if model_path is None:
            return
        self.model_path = model_path

    def text_detection_finished(self) -> None:
        """
        Close the dialog if all models have been downloaded.
        """
        self.text_detector_downloaded = True
        self.check_finished()

    # ============================= OCR =============================

    def download_ocr(self) -> None:
        """
        Download the ocr model.
        """
        # Start the download in a separate thread.
        threadpool = Qc.QThreadPool.globalInstance()

        custom_stdout = gu.CaptureOutput()
        custom_stdout.text_written.connect(self.ocr_output_log)

        worker = wt.Worker(
            ocr_download_worker,
            custom_stdout,
            no_progress_callback=True,
        )
        worker.signals.finished.connect(self.ocr_finished)
        worker.signals.error.connect(self.ocr_error)
        threadpool.start(worker)

    def ocr_finished(self) -> None:
        """
        Close the dialog if all models have been downloaded.
        """
        self.ocr_downloaded = True
        if self.buffer:
            logger.info("\n".join(self.buffer))
            gu.show_warning(self, "OCR download errors", "\n".join(self.buffer))
        self.check_finished()

    @Slot(wt.WorkerError)
    def ocr_error(self, worker_error: wt.WorkerError) -> None:
        """
        Show an error message.

        :param worker_error: The error message.
        """
        self.label_ocr_speed.setText("—")
        self.label_ocr_size.setText("")
        self.encountered_error = True
        logger.exception(worker_error.value)
        gu.show_warning(self, "Download Failed", str(worker_error.value))

    @Slot(str, str)
    def ocr_output_log(self, text: str, stream: str) -> None:
        """
        Append the text to the output log.

        :param text: The text to append.
        :param stream: The stream the text was written to, stdout or stderr.
        """
        progress_data = parse_ocr_progress(text)
        if progress_data is None:
            return
        if isinstance(progress_data, str):
            self.buffer.append(progress_data)
            return
        # Update the ui with the progress data.
        self.show_ocr_progress(progress_data)

    def show_ocr_progress(self, progress_data: ProgressData) -> None:
        self.label_ocr_file_name.setText(progress_data.file_name)
        self.label_ocr_size.setText(
            f"{format_size(progress_data.downloaded_size)} / {format_size(progress_data.file_size)}"
        )
        self.label_ocr_speed.setText(
            f"{format_size(progress_data.speed)}/s, ETA: {format_timespan(progress_data.eta, ) if progress_data.eta >= 0 else '—'}"
        )
        self.progressBar_ocr.setValue(progress_data.percentage)


def ocr_download_worker(custom_stdout) -> None:
    with custom_stdout:
        MangaOcr()


def download_file(url: str, save_dir: Path, signals: DownloadSignals, sha_hash: str = None) -> Path:
    """
    Download a file from a url.

    If something goes wrong, it will raise an exception.

    :param url: The url to download the file from.
    :param save_dir: The directory where the file will be saved.
    :param signals: The signals to emit progress and error messages.
    :param sha_hash: The sha256 hash of the file. If it doesn't match, the download will be aborted.
    :return: The path to the downloaded file.
    """
    response = requests.get(url, stream=True, timeout=10)
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise ConnectionError(f"Error downloading file from url: {url}\n{e}")

    file_name = os.path.basename(url)
    save_path = save_dir / file_name

    save_dir.mkdir(parents=True, exist_ok=True)
    if save_path.is_file():
        save_path.unlink()
    elif save_path.is_dir():
        save_path.rmdir()

    file_size = int(response.headers.get("Content-Length", 0))
    chunk_size = 8192  # 8KB
    downloaded_size = 0
    start_time = time.time()

    # Since we don't want to update the infos as often as the progress bar, only update it every few seconds.
    last_progress_time = start_time
    last_downloaded_size = 0
    last_speed = 0
    last_eta = -1

    with open(save_path, "wb") as f:
        for chunk in response.iter_content(chunk_size=chunk_size):
            if chunk:
                f.write(chunk)
                downloaded_size += len(chunk)

                elapsed_time = time.time() - start_time

                percentage = int((downloaded_size / file_size) * 100)

                # Only update the the rest of the infos if the time difference exceeds the update interval.
                if time.time() - last_progress_time > PROGRESS_UPDATE_INTERVAL or percentage == 100:
                    last_progress_time = time.time()
                    last_downloaded_size = downloaded_size
                    last_speed = downloaded_size / elapsed_time  # Bytes per second
                    remaining_bytes = file_size - downloaded_size
                    last_eta = round(remaining_bytes / last_speed)  # in seconds

                progress_data = ProgressData(
                    percentage=percentage,
                    downloaded_size=last_downloaded_size,
                    file_size=file_size,
                    speed=last_speed,
                    eta=last_eta,
                )
                signals.progress_signal.emit(progress_data)

    if not save_path.exists():
        raise OSError(
            f"Error downloading file from url: {url}\nFailed to save the file to {save_path}"
        )

    if sha_hash is not None:
        if not md.check_hash(save_path, sha_hash):
            save_path.unlink()
            raise OSError(
                f"Error downloading file from url: {url}\nThe file content is different from expected."
            )

    return save_path


def parse_ocr_progress(line: str) -> ProgressData | str | None:
    """
    Parse a line of the download progress. The huggingface transformer library ostensibly uses
    tqdm to show the download progress.
    Example: '\rDownloading pytorch_model.bin:  12%|#1        | 31.5M/444M [00:04<00:59, 6.95MB/s]'

    Note: The specific application I'm tracking, Huggingface's transformer library, uses decimal size prefixes for data.

    :param line:
    :return:
    """
    line = line.strip()
    if not line:
        return None

    # Try matching known file names in the line
    known_filenames = [
        "preprocessor_config.json",
        "tokenizer_config.json",
        "config.json",
        "special_tokens_map.json",
        "vocab.txt",
        "pytorch_model.bin",
    ]

    # Using regex to capture relevant data
    regex_pattern = r".*?Downloading (?P<path>.*?): *(?P<percentage>\d+)%\|.*?\| (?P<downloaded>[\d.]+[kMGT]?B?)/(?P<total>[\d.]+[kMGT]?B?)(?: \[.*?<(?P<eta>.*), (?P<speed>[\d.]+[kMGT]?B/s)\])?"

    match = re.search(regex_pattern, line)
    if not match:
        return "Could not parse progress data: " + line

    # Figure out the file name. If it was not elided, use the file name from the line.
    # Otherwise, check if there is a directory delimiter in the path and use the last part of the path as the file name.
    # If all else fails, try and match it to a known file name.
    filename = match.group("path")
    if filename.startswith("(…)"):
        filename = filename[3:]
        if "/" in filename:
            filename = filename.split("/")[-1]
        else:
            for known_filename in known_filenames:
                if known_filename in filename:
                    filename = known_filename
                    break

    else:
        filename = filename.split("/")[-1]

    percentage = int(match.group("percentage"))

    downloaded = parse_size(match.group("downloaded"))
    total = parse_size(match.group("total"))

    speed = parse_size(match.group("speed")) if match.group("speed") else 0

    eta = parse_simple_time(match.group("eta")) if match.group("eta") else -1

    return ProgressData(percentage, downloaded, total, speed, eta, filename)


def parse_simple_time(time_str: str) -> int:
    """
    Parse time formatted by tqdm to total seconds.
    The expected format: "00:00" or "00:00:00"

    :param time_str: Time string formatted by tqdm (e.g., "240:25:35").
    :return: Total seconds.
    :raises ValueError: If the time format is invalid.
    """
    # Split the time string by colons
    parts = time_str.split(":")

    if len(parts) == 2:
        # Minutes and seconds are provided.
        return int(parts[0]) * 60 + int(parts[1])
    elif len(parts) == 3:
        # Hours, minutes, and seconds are provided.
        return int(parts[0]) * 60 * 60 + int(parts[1]) * 60 + int(parts[2])
    else:
        raise ValueError("Invalid time format")



import io
from enum import Enum, IntEnum, auto
from pathlib import Path
from typing import Iterable
from typing import Protocol, Any
from uuid import uuid4

import PySide6.QtGui as Qg
import attrs
import dictdiffer
import humanfriendly
from PIL import Image
from attrs import frozen, fields

import pcleaner.config as cfg


# The max size used for the icon and large thumbnail.
THUMBNAIL_SIZE = 64


class Output(IntEnum):
    """
    These enums represent all the output images displayed in the gui.
    These are an IntEnum so that they can be compared by order.
    This way, we can determine the last step in a set of outputs.
    """

    input = auto()
    ai_mask = auto()
    raw_json = auto()

    initial_boxes = auto()
    final_boxes = auto()
    ocr = auto()
    clean_json = auto()

    box_mask = auto()
    cut_mask = auto()
    mask_layers = auto()
    mask_overlay = auto()
    final_mask = auto()
    isolated_text = auto()
    masked_image = auto()
    mask_data_json = auto()

    denoiser_mask = auto()
    denoised_image = auto()

    write_output = auto()  # This is only used for the progress bar.


class Step(IntEnum):
    """
    These enums represent all the steps in the image processing pipeline.
    """

    text_detection = 1
    preprocessor = auto()
    masker = auto()
    denoiser = auto()
    output = auto()


class ProgressType(Enum):
    """
    The type of progress reported:
    - Begin: The step started, so the progress bar should be reset.
    - Incremental: One (or more) tasks were completed.
    - Absolute: The progress bar should be set to a specific value.
    - End: The step finished, so the progress bar should be set to 100%.
    """

    start = auto()
    begin_step = auto()
    incremental = auto()
    absolute = auto()
    analyticsOCR = auto()
    analyticsMasker = auto()
    analyticsDenoiser = auto()
    outputOCR = auto()
    end = auto()
    aborted = auto()


@frozen
class ProgressData:
    """
    A callback to report progress to the gui.

    Consists of:
    - The total number of images to process.
    - The target output to reach.
    - The current step.
    - The progress type.
    - A value for the case of absolute progress, or how much to increment by for incremental progress, or whatever analytics.
    """

    total_images: int
    target_outputs: list[Output]
    current_step: Step
    progress_type: ProgressType
    value: Any = 1  # Default value for incremental progress. Otherwise usually an analytics struct.


class ProgressSignal(Protocol):
    def emit(self, data: ProgressData) -> None:
        pass


output_to_step: dict[Output, Step] = {
    Output.input: Step.text_detection,
    Output.ai_mask: Step.text_detection,
    Output.raw_json: Step.text_detection,
    Output.initial_boxes: Step.preprocessor,
    Output.final_boxes: Step.preprocessor,
    Output.ocr: Step.preprocessor,
    Output.clean_json: Step.preprocessor,
    Output.box_mask: Step.masker,
    Output.cut_mask: Step.masker,
    Output.mask_layers: Step.masker,
    Output.final_mask: Step.masker,
    Output.mask_overlay: Step.masker,
    Output.isolated_text: Step.masker,
    Output.masked_image: Step.masker,
    Output.mask_data_json: Step.masker,
    Output.denoiser_mask: Step.denoiser,
    Output.denoised_image: Step.denoiser,
    Output.write_output: Step.output,
}

# The final output representing each step.
# If this output is intact, consider the step complete.
step_to_output: dict[Step, tuple[Output, ...]] = {
    Step.text_detection: (Output.input, Output.ai_mask, Output.raw_json),
    Step.preprocessor: (Output.initial_boxes, Output.final_boxes, Output.ocr, Output.clean_json),
    Step.masker: (
        Output.box_mask,
        Output.cut_mask,
        Output.mask_layers,
        Output.mask_overlay,
        Output.isolated_text,
        Output.masked_image,
        Output.final_mask,
        Output.mask_data_json,
    ),
    Step.denoiser: (Output.denoiser_mask, Output.denoised_image),
    Step.output: (Output.write_output,),
}

# When checking which outputs are affected by a profile change, we can check
# from the last step to the first step, and stop when we find an output that is
# unaffected by the profile change.
# But these outputs are affected by changes that the ones after them are not,
# so we can't use that optimization.
OUTPUTS_WITH_INDEPENDENT_PROFILE_SENSITIVITY = (Output.mask_overlay,)


class ImageAnalyticCategory(Enum):
    """
    The category of analytics.
    """

    ocr_removed = auto()
    mask_failed = auto()
    mask_perfect = auto()
    denoised = auto()


class ImageAnalytics:
    """
    Bundle all the necessary info to display the minimal analytics for an image.
    These numbers are shown below their corresponding icon in the file table.
    total means the number of boxes or masks from which the failed etc. numbers are derived.
    This total can shrink as the image is processed, since if a box is thrown out by ocr, there
    won't be a corresponding mask for it etc., hence the seemingly redundant multiple totals.

    The analytics per category are stored as simple strings in the form "value/total".

    If the analytic value is blank, it should be hidden in the gui, since it's not relevant.
    These analytics are supposed to give a quick overview of how bad or good the image is, or where
    problems might be.
    """

    _data: dict[ImageAnalyticCategory, str]

    def __init__(self) -> None:
        self._data = {category: "" for category in ImageAnalyticCategory}

    def get_category(self, category: ImageAnalyticCategory) -> str:
        return self._data[category]

    def set_category(self, category: ImageAnalyticCategory, value: int, total: int) -> None:
        """
        Set the analytic value for the category as a string internally.
        If either value is 0, the analytic is reset.

        :param category: The category to set the value for.
        :param value: The value to set.
        :param total: The total to set.
        """
        if value == 0 or total == 0:
            self._data[category] = ""
        else:
            self._data[category] = f"{value}/{total}"


def get_output_representing_step(step: Step) -> Output:
    """
    Get the output representing the step, which is the last one in the pipeline.
    This is the json output for the first three steps, which forms the foundation for the
    next step to continue from.

    :param step: The step to get the output for.
    :return: The output representing the step.
    """
    return step_to_output[step][-1]


class ProcessOutput:
    """
    This class represents an output in the image processing pipeline.
    The sensitivity filter is used to determine which entries in the profile will affect
    this output.
    """

    description: str
    step_name: str | None
    output_name: str | None
    _path: Path | None = None
    _sensitivity_filter: attrs.filters
    _current_profile_dict: dict[str, Any] | None = None
    # This used to use a current_dict_checksum, using the json representation of the filtered profile
    # for the digest, but that proved unreliable despite the json exporter having sort keys enabled.
    # The result were phantom differences that didn't exist, so using dictdiffer and comparing the filtered
    # dicts directly was chosen instead, since it's more reliable than some other handcrafted json output.

    def __init__(
        self,
        description: str,
        step_name: str | None,
        output_name: str | None,
        profile_sensitivity: Iterable[attrs.Attribute] | None = None,
    ):
        """
        Init the process output.

        Set up the filter for the profile sensitivity.
        Pass in a list of attrs attributes like this:
            [fields(Profile).some_attribute, fields(Profile).some_other_attribute]

        If the profile sensitivity is None, the output will include all profile entries.

        :param description: A description of the output.
        :param step_name: [Optional] The name of the step that this output belongs to. If none is given, this
            output will be hidden.
        :param output_name: [Optional] The name of the output. Left blank when it's the only output for the step.
        :param profile_sensitivity: [Optional] A list of attributes that this output is sensitive to. If None, all
            attributes will be included.
        """
        self.description = description
        self.step_name = step_name
        self.output_name = output_name
        if profile_sensitivity is None:
            self._sensitivity_filter = None
        else:
            self._sensitivity_filter = attrs.filters.include(*profile_sensitivity)

    def update(self, path: Path, profile: cfg.Profile) -> None:
        """
        Update the path of the output and the checksum of the profile entries that this output is sensitive to.

        :param path: The path to the output.
        :param profile: The profile to calculate the checksum for.
        """
        self._path = path
        self.update_checksum(profile)

    @property
    def path(self) -> Path | None:
        return self._path

    def has_path(self) -> bool:
        return self.path is not None

    def reset(self) -> None:
        """
        Clear the path and checksum.
        """
        self._path = None
        self._current_profile_dict = None

    def update_checksum(self, profile: cfg.Profile) -> None:
        """
        Update the checksum of the profile entries that this output is sensitive to.

        :param profile: The profile to calculate the checksum for.
        """
        self._current_profile_dict = self._profile_check_dict(profile)

    def is_changed(self, profile: cfg.Profile) -> bool:
        """
        Check if the profile entries that this output is sensitive to have changed.

        :param profile: The profile to check.
        :return: True if the profile has changed, False otherwise.
        """
        return self._current_profile_dict is None or next(
            dictdiffer.diff(self._current_profile_dict, self._profile_check_dict(profile)), False
        )

    def is_unchanged(self, profile: cfg.Profile) -> bool:
        """
        Check if the profile entries that this output is sensitive to haven't changed.

        :param profile: The profile to check.
        :return: True if the profile hasn't changed, False otherwise.
        """
        return not self.is_changed(profile)

    def _profile_check_dict(self, profile: cfg.Profile) -> dict:
        """
        Trim down the profile to only include entries that this output is sensitive to.

        :param profile: The profile to trim.
        :return: The trimmed profile.
        """
        profile_dict = attrs.asdict(profile, filter=self._sensitivity_filter, recurse=True)
        return profile_dict


class ImageFile:
    """
    This class represents an image file.
    """

    path: Path  # Path to the image file.
    icon: Qg.QIcon  # Placeholder icon for the image type.
    uuid: str  # Unique identifier for the temp files to prevent name collisions.
    # The following attributes are lazy-loaded.
    thumbnail: Qg.QPixmap | None = None  # Thumbnail of the image, used as the icon.
    size: tuple[int, int] | None = None  # Size of the image.
    file_size: int | None = None  # Size of the image file in bytes.
    color_mode: str | None = None  # Color mode of the image.
    outputs: dict[Output, ProcessOutput]  # Map of steps to ProcessStep objects.
    loading_queued: bool = False  # Whether the image is queued to be loaded.
    analytics_data: ImageAnalytics = None

    error: Exception | None = None  # Error that occurred during any process.

    def __init__(self, path: Path) -> None:
        """
        Init the image file.

        :param path: Path to the image file.
        """
        self.path = path
        self.icon = Qg.QIcon.fromTheme(cfg.SUFFIX_TO_ICON[path.suffix.lower()])
        self.uuid = str(uuid4())
        self.outputs = {}
        self.analytics_data = ImageAnalytics()

        pro = fields(cfg.Profile)
        gen = fields(cfg.GeneralConfig)
        td = fields(cfg.TextDetectorConfig)
        pp = fields(cfg.PreprocessorConfig)
        mk = fields(cfg.MaskerConfig)
        dn = fields(cfg.DenoiserConfig)

        # Init the process steps.
        # Here I need to account for all the settings that affect each step.
        # The settings are passed as a list of attributes, note that the parent attribute
        # from the profile must be included, for child attributes to be included.

        # Text Detection:
        settings = [pro.general, gen.input_height_lower_target, gen.input_height_upper_target]
        self.outputs[Output.input] = ProcessOutput(
            "The original image with scaling applied (if needed).", "Input", None, settings
        )
        settings += [pro.text_detector, td.model_path]
        self.outputs[Output.ai_mask] = ProcessOutput(
            "The rough mask generated by the AI.", "Text Detection", None, settings
        )
        self.outputs[Output.raw_json] = ProcessOutput("Not visible", None, None, settings)

        # Preprocessor:
        settings += [
            pro.preprocessor,
            pp.box_min_size,
            pp.suspicious_box_min_size,
            pp.box_padding_initial,
            pp.box_right_padding_initial,
        ]

        self.outputs[Output.initial_boxes] = ProcessOutput(
            "The outlines of the text boxes the AI found.",
            "Preprocessor",
            "Initial Boxes",
            settings,
        )
        settings += [
            pp.ocr_enabled,
            pp.ocr_max_size,
            pp.ocr_blacklist_pattern,
            pp.box_padding_extended,
            pp.box_right_padding_extended,
            pp.box_reference_padding,
        ]
        self.outputs[Output.final_boxes] = ProcessOutput(
            "The final boxes after expanding, merging and filtering unneeded boxes with OCR.\n"
            "Green: initial boxes. Red: extended boxes. Purple: merged (final) boxes. "
            "Blue: reference boxes for denoising.",
            "Preprocessor",
            "Final Boxes",
            settings,
        )
        self.outputs[Output.clean_json] = ProcessOutput("Not visible", None, None, settings)

        # Masker:
        self.outputs[Output.box_mask] = ProcessOutput(
            "The mask of the merged boxes.", "Masker", "Box Mask", settings
        )
        self.outputs[Output.cut_mask] = ProcessOutput(
            "The rough text detection mask with everything outside the box mask cut out.",
            "Masker",
            "Cut Mask",
            settings,
        )
        settings += [pro.masker, mk.mask_growth_step_pixels, mk.mask_growth_steps]
        self.outputs[Output.mask_layers] = ProcessOutput(
            "The different steps of growth around the cut mask displayed in different colors.",
            "Masker",
            "Mask Layers",
            settings,
        )
        settings += [
            mk.off_white_max_threshold,
            mk.mask_improvement_threshold,
            mk.mask_selection_fast,
            mk.mask_max_standard_deviation,
        ]
        self.outputs[Output.mask_overlay] = ProcessOutput(
            "The input image with the final mask overlaid in color.",
            "Masker",
            "Mask Overlay",
            settings + [mk.debug_mask_color],
        )
        self.outputs[Output.final_mask] = ProcessOutput(
            "The collection of masks for each bubble that fit best.",
            "Masker",
            "Final Mask",
            settings,
        )
        self.outputs[Output.isolated_text] = ProcessOutput(
            "The text layer isolated from the input image.", "Masker", "Isolated Text", settings
        )
        self.outputs[Output.masked_image] = ProcessOutput(
            "The input image with the final mask applied.", "Masker", "Masked Output", settings
        )
        self.outputs[Output.mask_data_json] = ProcessOutput("Not visible", None, None, settings)

        # Denoiser:
        settings += [
            pro.denoiser,
            dn.denoising_enabled,
            dn.noise_min_standard_deviation,
            dn.noise_outline_size,
            dn.noise_fade_radius,
            dn.colored_images,
            dn.filter_strength,
            dn.color_filter_strength,
            dn.template_window_size,
            dn.search_window_size,
        ]
        self.outputs[Output.denoiser_mask] = ProcessOutput(
            "The masks that required denoising, to be overlaid on the final mask when exporting.",
            "Denoiser",
            "Denoise Mask",
            settings,
        )
        self.outputs[Output.denoised_image] = ProcessOutput(
            "The input image with the denoised mask applied.",
            "Denoiser",
            "Denoised Output",
            settings,
        )

    @property
    def size_str(self) -> str:
        if self.size is None:
            return "Unknown"
        return f"{self.size[0]:n} × {self.size[1]:n}"

    @property
    def file_size_str(self) -> str:
        if self.file_size is None:
            return "Unknown"
        return humanfriendly.format_size(self.file_size, binary=True)

    @property
    def color_mode_str(self) -> str:
        if self.color_mode is None:
            return "Unknown"
        elif self.color_mode in ("RGB", "RGBA"):
            return "RGB"
        elif self.color_mode == "CMYK":
            return "CMYK"
        elif self.color_mode in ("L", "LA"):
            return "Grayscale"
        elif self.color_mode == "1":
            return "1-bit"
        elif self.color_mode == "P":
            return "Palette"
        else:
            return "Unknown"

    def data_loaded(self) -> bool:
        """
        Returns whether the lazily-loaded image data is ready.
        """
        return self.thumbnail is not None and self.size is not None

    def load_image(self) -> Path:
        """
        Loads the image data.

        :return: Path to the image file so that the callback knows which image was loaded.
        """
        image = Image.open(self.path)
        self.size = image.size
        self.file_size = self.path.stat().st_size
        self.color_mode = image.mode
        # Shrink the image down to a thumbnail size to reduce memory usage.
        thumbnail = image.copy()
        thumbnail.thumbnail((THUMBNAIL_SIZE, THUMBNAIL_SIZE))
        self.thumbnail = convert_PIL_to_QPixmap(thumbnail)

        return self.path


def convert_PIL_to_QPixmap(image: Image.Image) -> Qg.QPixmap:
    """
    Converts a PIL image to a QPixmap.

    :param image: PIL image.
    :return: QPixmap.
    """
    if image.mode == "CMYK":
        image = image.convert("RGB")
    byte_data = io.BytesIO()
    image.save(byte_data, format="PNG")
    byte_data.seek(0)
    qimage = Qg.QPixmap()
    # noinspection PyTypeChecker
    qimage.loadFromData(byte_data.read(), "PNG")
    return qimage



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'About.ui'
##
## Created by: Qt User Interface Compiler version 6.5.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFormLayout, QHBoxLayout, QLabel,
    QSizePolicy, QSpacerItem, QVBoxLayout, QWidget)

class Ui_About(object):
    def setupUi(self, About):
        if not About.objectName():
            About.setObjectName(u"About")
        About.resize(600, 240)
        self.horizontalLayout_2 = QHBoxLayout(About)
        self.horizontalLayout_2.setSpacing(6)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(20, 20, 20, 20)
        self.label_logo = QLabel(About)
        self.label_logo.setObjectName(u"label_logo")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_logo.sizePolicy().hasHeightForWidth())
        self.label_logo.setSizePolicy(sizePolicy)
        self.label_logo.setMinimumSize(QSize(200, 200))
        self.label_logo.setMaximumSize(QSize(200, 200))
        self.label_logo.setPixmap(QPixmap(u":/logo.png"))
        self.label_logo.setScaledContents(True)

        self.horizontalLayout_2.addWidget(self.label_logo)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Fixed, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label_2 = QLabel(About)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setTextFormat(Qt.MarkdownText)

        self.verticalLayout.addWidget(self.label_2)

        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setContentsMargins(-1, 10, -1, -1)
        self.label_3 = QLabel(About)
        self.label_3.setObjectName(u"label_3")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label_3)

        self.label_version = QLabel(About)
        self.label_version.setObjectName(u"label_version")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.label_version)

        self.label_11 = QLabel(About)
        self.label_11.setObjectName(u"label_11")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_11)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_copyright = QLabel(About)
        self.label_copyright.setObjectName(u"label_copyright")

        self.horizontalLayout.addWidget(self.label_copyright)

        self.label_13 = QLabel(About)
        self.label_13.setObjectName(u"label_13")

        self.horizontalLayout.addWidget(self.label_13)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.formLayout.setLayout(1, QFormLayout.FieldRole, self.horizontalLayout)

        self.label_5 = QLabel(About)
        self.label_5.setObjectName(u"label_5")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.label_5)

        self.label_toolkit = QLabel(About)
        self.label_toolkit.setObjectName(u"label_toolkit")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.label_toolkit)

        self.label_7 = QLabel(About)
        self.label_7.setObjectName(u"label_7")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.label_7)

        self.label_license = QLabel(About)
        self.label_license.setObjectName(u"label_license")
        self.label_license.setTextInteractionFlags(Qt.LinksAccessibleByKeyboard|Qt.LinksAccessibleByMouse)

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.label_license)

        self.label_9 = QLabel(About)
        self.label_9.setObjectName(u"label_9")

        self.formLayout.setWidget(4, QFormLayout.LabelRole, self.label_9)

        self.label_10 = QLabel(About)
        self.label_10.setObjectName(u"label_10")
        self.label_10.setOpenExternalLinks(True)
        self.label_10.setTextInteractionFlags(Qt.LinksAccessibleByKeyboard|Qt.LinksAccessibleByMouse)

        self.formLayout.setWidget(4, QFormLayout.FieldRole, self.label_10)


        self.verticalLayout.addLayout(self.formLayout)


        self.horizontalLayout_2.addLayout(self.verticalLayout)


        self.retranslateUi(About)

        QMetaObject.connectSlotsByName(About)
    # setupUi

    def retranslateUi(self, About):
        About.setWindowTitle(QCoreApplication.translate("About", u"About Panel Cleaner", None))
        self.label_logo.setText("")
        self.label_2.setText(QCoreApplication.translate("About", u"# Panel Cleaner\n"
"\n"
"An AI-powered tool to clean manga panels.", None))
        self.label_3.setText(QCoreApplication.translate("About", u"Version:", None))
        self.label_version.setText(QCoreApplication.translate("About", u"<version>", None))
        self.label_11.setText(QCoreApplication.translate("About", u"Author:", None))
        self.label_copyright.setText(QCoreApplication.translate("About", u"<copyright>", None))
        self.label_13.setText(QCoreApplication.translate("About", u"Voxel", None))
        self.label_5.setText(QCoreApplication.translate("About", u"Toolkit:", None))
        self.label_toolkit.setText(QCoreApplication.translate("About", u"<toolkit>", None))
        self.label_7.setText(QCoreApplication.translate("About", u"License:", None))
        self.label_license.setText(QCoreApplication.translate("About", u"<html><head/><body><p><a href=\"open_license\"><span style=\" text-decoration: underline; color:#8419ff;\">GNU General Public License v3.0</span></a></p></body></html>", None))
        self.label_9.setText(QCoreApplication.translate("About", u"Source code:", None))
        self.label_10.setText(QCoreApplication.translate("About", u"<html><head/><body><p><a href=\"https://github.com/VoxelCubes/PanelCleaner\"><span style=\" text-decoration: underline; color:#8419ff;\">GitHub</span></a></p></body></html>", None))
    # retranslateUi




import itertools
from copy import deepcopy
from functools import partial
from io import StringIO
from multiprocessing import Pool
from pathlib import Path
from typing import Callable

import torch
from PIL import Image
from loguru import logger
from manga_ocr import MangaOcr
from natsort import natsorted

import pcleaner.config as cfg
import pcleaner.denoiser as dn
import pcleaner.gui.ctd_interface_gui as ctm
import pcleaner.gui.image_file as imf
import pcleaner.gui.worker_thread as wt
import pcleaner.helpers as hp
import pcleaner.image_ops as ops
import pcleaner.masker as ma
import pcleaner.preprocessor as pp
import pcleaner.structures as st


def generate_output(
    image_objects: list[imf.ImageFile],
    target_outputs: list[imf.Output],
    output_dir: Path | None,
    config: cfg.Config,
    ocr_model: MangaOcr | None,
    progress_callback: imf.ProgressSignal,
    abort_flag: wt.SharableFlag,
    debug: bool = False,
):
    """
    Process all the images in the given image objects list until the target step is reached.
    The paths inside the image objects are updated as the processing progresses.
    If the profile checksum of a step hasn't changed, the step is skipped.

    In GUI mode, the masker and denoiser never write the output directly, but instead
    place them in the cache directory. From there they may be copied to the output directory,
    to ensure that effort isn't duplicated where possible. For example, if the output was already
    generated for the viewer, it doesn't need to be generated again for the final output, or when
    only changing the output path for the final output.

    The ocr model (or None) is passed in as arguments, so it can be reused
    across multiple runs. This is done because in gui mode the model does not need to be loaded
    repeatedly for single image runs, making iterations much faster, due to the slow
    initializations of the model taking several seconds.

    When setting the output dir to None, intermediate steps are saved to disk, offering
    additional visualizations that aren't necessary when you aren't viewing image details.

    :param image_objects: The image objects to process. These contain the image paths.
    :param target_outputs: The target outputs to reach, eg. the clean mask and image.
        Multiple outputs are only supported for those within the same step, as it's meant to accommodate
        the full run parameters.
    :param output_dir: The directory to save the outputs to. If None, the output remains in the cache directory.
    :param config: The config to use.
    :param ocr_model: The ocr model to use.
    :param progress_callback: A callback to report progress to the gui.
    :param abort_flag: A flag that is set to True when the thread should abort.
    :param debug: If True, debug messages are printed.
    """

    # We will check before each step and during each iteration of the step to see if the thread should abort.
    # The check is NOT done right before checking the outputs to avoid throwing away perfectly fine work.
    # This means that the abort check shouldn't interrupt the processing step on an image, leaving it in a
    # dirty state.
    def check_abortion() -> None:
        """
        Check if the thread should abort.
        If so, signal abortion to the progress callback and raise an AbortException.
        """
        if abort_flag.get():
            progress_callback.emit(
                imf.ProgressData(
                    0,
                    [],
                    imf.Step.output,
                    imf.ProgressType.aborted,
                )
            )
            raise wt.Abort()

    progress_callback.emit(
        imf.ProgressData(
            0,
            [],
            imf.Step.text_detection,
            imf.ProgressType.start,
        )
    )

    # Create an independant copy of the profile, since this instance is shared and mutated across threads.
    # Due to the checksum only being generated at the end, we don't want to erroneously claim this output
    # matches a profile that was changed in the meantime.
    profile = deepcopy(config.current_profile)

    cache_dir = config.get_cleaner_cache_dir()

    # Pick out the highest output we need to reach.
    # Other outputs are only considered if they are in the same step, used for final output runs.
    target_output = max(target_outputs)

    # Get the text detector model path from the config.
    cuda = torch.cuda.is_available()
    text_detector_model_path = config.get_model_path(cuda)

    def step_needs_to_be_rerun_closure(current_step: imf.Step) -> Callable[[imf.ImageFile], bool]:
        """
        Check if any of the outputs for the given image object need to be rerun.

        :return: A function that takes an image object and returns True if the step needs to be rerun for it.
        """

        def step_changed(image_object: imf.ImageFile) -> bool:
            nonlocal profile
            nonlocal target_outputs
            nonlocal current_step
            # holy shit this is cursed I need to think about it more.
            # I think I got it sorted now...
            # Pretty sure it works :)
            # It didn't.
            # Now it is fr fr

            target_outputs_in_the_current_step: tuple[imf.Output] = tuple(
                filter(lambda o: o in imf.step_to_output[current_step], reversed(target_outputs))
            )

            if target_outputs_in_the_current_step:
                # We need to check each output we care about to know if we gotta rerun the step.
                return any(
                    image_object.outputs[o].is_changed(profile)
                    for o in target_outputs_in_the_current_step
                )
            else:
                # In this case we just need to know if the step's representative is complete.
                representative: imf.Output = imf.get_output_representing_step(current_step)
                return image_object.outputs[representative].is_changed(profile)

        return step_changed

    def update_output(image_object: imf.ImageFile, output: imf.Output, suffix: str) -> None:
        """
        Update the output of the given image object.
        Check if the file actually exists, and if it does, update the output path.

        :param image_object: The image object to update.
        :param output: The output to update.
        :param suffix: The suffix to add to the output path.
        """
        nonlocal profile, cache_dir

        path = cache_dir / f"{image_object.uuid}_{image_object.path.stem}{suffix}"

        if path.is_file():
            image_object.outputs[output].update(path, profile)

    # ============================================== Text Detection ==============================================

    check_abortion()

    step_text_detector_images = tuple(
        filter(step_needs_to_be_rerun_closure(imf.Step.text_detection), image_objects)
    )

    if step_text_detector_images:
        logger.info(
            f"Running text detection AI model for {len(step_text_detector_images)} images..."
        )
        try:
            ctm.model2annotations_gui(
                profile.general,
                profile.text_detector,
                text_detector_model_path,
                step_text_detector_images,
                cache_dir,
                no_text_detection=target_output == imf.Output.input,
                partial_progress_data=partial(
                    imf.ProgressData,
                    len(step_text_detector_images),
                    target_outputs,
                    imf.Step.text_detection,
                ),
                progress_callback=progress_callback if target_output != imf.Output.input else None,
                abort_flag=abort_flag,
            )
        except wt.Abort:
            # Send the progress callback a signal that the process was aborted,
            # and then raise the exception again to stop the processing.
            check_abortion()

        # Update the outputs of the image objects.
        for image_obj in step_text_detector_images:
            update_output(image_obj, imf.Output.input, ".png")
            update_output(image_obj, imf.Output.ai_mask, "_mask.png")
            update_output(image_obj, imf.Output.raw_json, "#raw.json")

    # ============================================== Preprocessing ==============================================

    check_abortion()

    if target_output > imf.get_output_representing_step(imf.Step.text_detection):
        step_preprocessor_images = tuple(
            filter(step_needs_to_be_rerun_closure(imf.Step.preprocessor), image_objects)
        )

        if step_preprocessor_images:
            logger.info(f"Running preprocessing for {len(step_preprocessor_images)} images...")

            progress_callback.emit(
                imf.ProgressData(
                    len(step_preprocessor_images),
                    target_outputs,
                    imf.Step.preprocessor,
                    imf.ProgressType.begin_step,
                )
            )

            ocr_analytics: list[st.OCRAnalytic] = []
            # Find all the json files associated with the images.
            for image_obj in step_preprocessor_images:
                check_abortion()
                json_file_path = cache_dir / f"{image_obj.uuid}_{image_obj.path.stem}#raw.json"
                ocr_analytic = pp.prep_json_file(
                    json_file_path,
                    preprocessor_conf=profile.preprocessor,
                    cache_masks=target_output in imf.step_to_output[imf.Step.preprocessor]
                    or output_dir is None,
                    mocr=ocr_model if profile.preprocessor.ocr_enabled else None,
                )

                if ocr_analytic is not None:
                    ocr_analytics.append(ocr_analytic)

                progress_callback.emit(
                    imf.ProgressData(
                        len(step_preprocessor_images),
                        target_outputs,
                        imf.Step.preprocessor,
                        imf.ProgressType.incremental,
                    )
                )

            # Update the outputs of the image objects.
            for image_obj in step_preprocessor_images:
                update_output(image_obj, imf.Output.initial_boxes, "_boxes.png")
                update_output(image_obj, imf.Output.final_boxes, "_boxes_final.png")
                update_output(image_obj, imf.Output.clean_json, "#clean.json")

            progress_callback.emit(
                imf.ProgressData(
                    len(step_preprocessor_images),
                    target_outputs,
                    imf.Step.preprocessor,
                    imf.ProgressType.analyticsOCR,
                    (ocr_analytics, profile.preprocessor.ocr_max_size),
                )
            )

    # ============================================== Masker ==============================================

    check_abortion()

    if target_output > imf.get_output_representing_step(imf.Step.preprocessor):
        step_masker_images = tuple(
            filter(step_needs_to_be_rerun_closure(imf.Step.masker), image_objects)
        )

        if step_masker_images:
            logger.info(f"Running masker for {len(step_masker_images)} images...")

            progress_callback.emit(
                imf.ProgressData(
                    len(step_masker_images),
                    target_outputs,
                    imf.Step.masker,
                    imf.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = (
                cache_dir / f"{image_obj.uuid}_{image_obj.path.stem}#clean.json"
                for image_obj in step_masker_images
            )

            # Pack all the arguments into a dataclass.
            outputs_that_need_masks = (
                imf.Output.box_mask,
                imf.Output.cut_mask,
                imf.Output.mask_layers,
                imf.Output.mask_overlay,
            )
            # If any of out targeted outputs need to have intermediate masks generated,
            # we need to show the masks. Or if we explicitly want to show the masks.
            need_to_show_masks = output_dir is None or any(
                output in outputs_that_need_masks for output in target_outputs
            )

            data = [
                st.MaskerData(
                    json_file,
                    None,
                    cache_dir,
                    profile.general,
                    profile.masker,
                    save_only_mask=target_outputs == [imf.Output.final_mask],
                    save_only_cleaned=target_outputs == [imf.Output.masked_image],
                    save_only_text=target_outputs == [imf.Output.isolated_text],
                    extract_text=imf.Output.isolated_text in target_outputs,
                    show_masks=need_to_show_masks,
                    debug=debug,
                )
                for json_file in json_files
            ]

            masker_analytics_raw = []

            if len(step_masker_images) > 2:
                # Only use multiprocessing if there are more than 2 images.
                with Pool() as pool:
                    for analytic in pool.imap(ma.clean_page, data):
                        check_abortion()
                        masker_analytics_raw.extend(analytic)

                        progress_callback.emit(
                            imf.ProgressData(
                                len(step_masker_images),
                                target_outputs,
                                imf.Step.masker,
                                imf.ProgressType.incremental,
                            )
                        )
            else:
                for data_obj in data:
                    check_abortion()
                    analytic = ma.clean_page(data_obj)
                    masker_analytics_raw.extend(analytic)

                    progress_callback.emit(
                        imf.ProgressData(
                            len(step_masker_images),
                            target_outputs,
                            imf.Step.masker,
                            imf.ProgressType.incremental,
                        )
                    )

            # Update the outputs of the image objects.
            for image_obj in step_masker_images:
                update_output(image_obj, imf.Output.box_mask, "_box_mask.png")
                update_output(image_obj, imf.Output.cut_mask, "_cut_mask.png")
                update_output(image_obj, imf.Output.mask_layers, "_masks.png")
                update_output(image_obj, imf.Output.final_mask, "_combined_mask.png")
                update_output(image_obj, imf.Output.mask_overlay, "_with_masks.png")
                update_output(image_obj, imf.Output.isolated_text, "_text.png")
                update_output(image_obj, imf.Output.masked_image, "_clean.png")
                update_output(image_obj, imf.Output.mask_data_json, "#mask_data.json")

            progress_callback.emit(
                imf.ProgressData(
                    len(step_masker_images),
                    target_outputs,
                    imf.Step.masker,
                    imf.ProgressType.analyticsMasker,
                    masker_analytics_raw,
                )
            )

    # ============================================== Denoiser ==============================================

    check_abortion()

    if (
        target_output > imf.get_output_representing_step(imf.Step.masker)
        and profile.denoiser.denoising_enabled
    ):
        step_denoiser_images = tuple(
            filter(step_needs_to_be_rerun_closure(imf.Step.denoiser), image_objects)
        )

        if step_denoiser_images:
            logger.info(f"Running denoiser for {len(step_denoiser_images)} images...")

            progress_callback.emit(
                imf.ProgressData(
                    len(step_denoiser_images),
                    target_outputs,
                    imf.Step.denoiser,
                    imf.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = (
                cache_dir / f"{image_obj.uuid}_{image_obj.path.stem}#mask_data.json"
                for image_obj in step_denoiser_images
            )

            # Pack all the arguments into a dataclass.
            data = [
                st.DenoiserData(
                    json_file,
                    None,
                    cache_dir,
                    profile.general,
                    profile.denoiser,
                    save_only_mask=target_outputs == [imf.Output.denoiser_mask],
                    save_only_cleaned=target_outputs == [imf.Output.denoised_image],
                    extract_text=imf.Output.isolated_text in target_outputs,
                    separate_noise_masks=False,
                    show_masks=True,
                    debug=debug,
                )
                for json_file in json_files
            ]

            denoise_analytics_raw = []
            if len(step_denoiser_images) > 2:
                with Pool() as pool:
                    for analytic in pool.imap(dn.denoise_page, data):
                        check_abortion()
                        denoise_analytics_raw.append(analytic)

                        progress_callback.emit(
                            imf.ProgressData(
                                len(step_denoiser_images),
                                target_outputs,
                                imf.Step.denoiser,
                                imf.ProgressType.incremental,
                            )
                        )
            else:
                for data_obj in data:
                    check_abortion()
                    analytic = dn.denoise_page(data_obj)
                    denoise_analytics_raw.append(analytic)

                    progress_callback.emit(
                        imf.ProgressData(
                            len(step_denoiser_images),
                            target_outputs,
                            imf.Step.denoiser,
                            imf.ProgressType.incremental,
                        )
                    )

            # Update the outputs of the image objects.
            for image_obj in step_denoiser_images:
                update_output(image_obj, imf.Output.denoiser_mask, "_noise_mask.png")
                update_output(image_obj, imf.Output.denoised_image, "_clean_denoised.png")
                update_output(image_obj, imf.Output.isolated_text, "_text.png")

            progress_callback.emit(
                imf.ProgressData(
                    len(step_denoiser_images),
                    target_outputs,
                    imf.Step.denoiser,
                    imf.ProgressType.analyticsDenoiser,
                    (
                        denoise_analytics_raw,
                        profile.denoiser.noise_min_standard_deviation,
                        profile.masker.mask_max_standard_deviation,
                    ),
                )
            )

    logger.info(f"Finished processing {len(image_objects)} images.")

    # ============================================== Final Output ==============================================

    check_abortion()

    if target_output == imf.Output.write_output:
        progress_callback.emit(
            imf.ProgressData(
                len(image_objects),
                target_outputs,
                imf.Step.output,
                imf.ProgressType.begin_step,
            )
        )

        for image_obj in image_objects:
            check_abortion()
            copy_to_output(image_obj, target_outputs, output_dir, profile)

            progress_callback.emit(
                imf.ProgressData(
                    len(image_objects),
                    target_outputs,
                    imf.Step.output,
                    imf.ProgressType.incremental,
                )
            )

    progress_callback.emit(
        imf.ProgressData(
            0,
            [],
            imf.Step.output,
            imf.ProgressType.end,
        )
    )


def copy_to_output(
    image_object: imf.ImageFile,
    outputs: list[imf.Output],
    output_directory: Path,
    profile: cfg.Profile,
):
    """
    Copy or export the outputs from the cache directory to the output directory.
    Output paths and preferred file types are taken into account.

    Supported outputs:
    - Masked image: Output.masked_image
    - Final mask: Output.final_mask
    - Isolated text: Output.isolated_text
    - Denoised image: Output.denoised_image
    - Denoised Mask: Output.denoiser_mask

    This may raise OSError in various circumstances.

    :param image_object: The image object to copy the outputs for.
    :param outputs: The outputs to copy.
    :param output_directory: The directory to write the outputs to.
    :param profile: The profile to use.
    """

    if output_directory.is_absolute():
        # When absolute, the output directory is used as is.
        final_out_path = output_directory / image_object.path.name
    else:
        # Otherwise, the output directory is relative to the original image's parent directory.
        final_out_path = image_object.path.parent / output_directory / image_object.path.name

    final_out_path.parent.mkdir(parents=True, exist_ok=True)

    # Create png paths for the outputs.
    cleaned_out_path = final_out_path.with_name(final_out_path.stem + "_clean.png")
    masked_out_path = final_out_path.with_name(final_out_path.stem + "_mask.png")
    text_out_path = final_out_path.with_name(final_out_path.stem + "_text.png")

    # Check what the preferred output format is.
    if profile.general.preferred_file_type is None:
        # Use the original file type by default.
        cleaned_out_path = cleaned_out_path.with_suffix(image_object.path.suffix)
    else:
        cleaned_out_path = cleaned_out_path.with_suffix(profile.general.preferred_file_type)

    if profile.general.preferred_mask_file_type is None:
        # Use png by default.
        masked_out_path = masked_out_path.with_suffix(".png")
        text_out_path = text_out_path.with_suffix(".png")
    else:
        masked_out_path = masked_out_path.with_suffix(profile.general.preferred_mask_file_type)
        text_out_path = text_out_path.with_suffix(profile.general.preferred_mask_file_type)

    # Output optimized images for all requested outputs.
    if imf.Output.masked_image in outputs:
        ops.save_optimized(
            image_object.outputs[imf.Output.masked_image].path, cleaned_out_path, image_object.path
        )

    if imf.Output.final_mask in outputs:
        # First scale the output mask to the original image size.
        final_mask = Image.open(image_object.outputs[imf.Output.final_mask].path)
        final_mask.resize(image_object.size, Image.NEAREST)
        ops.save_optimized(final_mask, masked_out_path)

    if imf.Output.isolated_text in outputs:
        ops.save_optimized(image_object.outputs[imf.Output.isolated_text].path, text_out_path)

    if imf.Output.denoised_image in outputs:
        ops.save_optimized(
            image_object.outputs[imf.Output.denoised_image].path,
            cleaned_out_path,
            image_object.path,
        )

    if imf.Output.denoiser_mask in outputs:
        # Special case: Here we need to take the final mask and paste this on top.
        final_mask = Image.open(image_object.outputs[imf.Output.final_mask].path)
        denoised_mask = Image.open(image_object.outputs[imf.Output.denoiser_mask].path)
        final_mask.paste(denoised_mask, (0, 0), denoised_mask)
        ops.save_optimized(final_mask, masked_out_path)


def perform_ocr(
    image_objects: list[imf.ImageFile],
    output_file: Path | None,
    csv_output: bool,
    config: cfg.Config,
    ocr_model: MangaOcr | None,
    progress_callback: imf.ProgressSignal,
    abort_flag: wt.SharableFlag,
    debug: bool = False,
):
    """
    Perform OCR on all the images in the given image objects list.
    If a path is given, the output is written to that path.
    Output is either a plain text file or csv.

    The ocr model (or None) is passed in as arguments, so it can be reused
    across multiple runs. This is done because in gui mode the model does not need to be loaded
    repeatedly for single image runs, making iterations much faster, due to the slow
    initializations of the model taking several seconds.

    When setting the output dir to None, intermediate steps are saved to disk, offering
    additional visualizations that aren't necessary when you aren't viewing image details.

    :param image_objects: The image objects to process. These contain the image paths.
    :param output_file: The directory to save the outputs to. If None, the output remains in the cache directory.
    :param csv_output: If True, the output is written as a csv file.
    :param config: The config to use.
    :param ocr_model: The ocr model to use.
    :param progress_callback: A callback to report progress to the gui.
    :param abort_flag: A flag that is set to True when the thread should abort.
    :param debug: If True, debug messages are printed.
    """

    def check_abortion() -> None:
        """
        Check if the thread should abort.
        If so, signal abortion to the progress callback and raise an AbortException.
        """
        if abort_flag.get():
            progress_callback.emit(
                imf.ProgressData(
                    0,
                    [],
                    imf.Step.output,
                    imf.ProgressType.aborted,
                )
            )
            raise wt.Abort()

    target_outputs = [imf.Output.ocr]

    progress_callback.emit(
        imf.ProgressData(
            0,
            [],
            imf.Step.text_detection,
            imf.ProgressType.start,
        )
    )

    # Create an independant copy of the profile, since this instance is shared and mutated across threads.
    # Due to the checksum only being generated at the end, we don't want to erroneously claim this output
    # matches a profile that was changed in the meantime.
    # Also, we will be editing this profile to get OCR working.
    profile = deepcopy(config.current_profile)

    # Make sure OCR is enabled.
    profile.preprocessor.ocr_enabled = True
    # Make sure the max size is infinite, so no boxes are skipped in the OCR process.
    profile.preprocessor.ocr_max_size = 10**10
    # Make sure the sus box min size is infinite, so all boxes with "unknown" language are skipped.
    profile.preprocessor.suspicious_box_min_size = 10**10
    # Set the OCR blacklist pattern to match everything, so all text gets reported in the analytics.
    profile.preprocessor.ocr_blacklist_pattern = ".*"

    cache_dir = config.get_cleaner_cache_dir()

    # Get the text detector model path from the config.
    cuda = torch.cuda.is_available()
    text_detector_model_path = config.get_model_path(cuda)

    def step_needs_to_be_rerun_closure(current_step: imf.Step) -> Callable[[imf.ImageFile], bool]:
        """
        Check if any of the outputs for the given image object need to be rerun.

        :return: A function that takes an image object and returns True if the step needs to be rerun for it.
        """

        def step_changed(image_object: imf.ImageFile) -> bool:
            nonlocal profile
            nonlocal target_outputs
            nonlocal current_step

            target_outputs_in_the_current_step: tuple[imf.Output] = tuple(
                filter(lambda o: o in imf.step_to_output[current_step], reversed(target_outputs))
            )

            if target_outputs_in_the_current_step:
                # We need to check each output we care about to know if we gotta rerun the step.
                return any(
                    image_object.outputs[o].is_changed(profile)
                    for o in target_outputs_in_the_current_step
                )
            else:
                # In this case we just need to know if the step's representative is complete.
                representative: imf.Output = imf.get_output_representing_step(current_step)
                return image_object.outputs[representative].is_changed(profile)

        return step_changed

    def update_output(image_object: imf.ImageFile, output: imf.Output, suffix: str) -> None:
        """
        Update the output of the given image object.
        Check if the file actually exists, and if it does, update the output path.

        :param image_object: The image object to update.
        :param output: The output to update.
        :param suffix: The suffix to add to the output path.
        """
        nonlocal profile, cache_dir

        _path = cache_dir / f"{image_object.uuid}_{image_object.path.stem}{suffix}"

        if _path.is_file():
            image_object.outputs[output].update(_path, profile)

    # ============================================== Text Detection ==============================================

    check_abortion()

    step_text_detector_images = tuple(
        filter(step_needs_to_be_rerun_closure(imf.Step.text_detection), image_objects)
    )

    if step_text_detector_images:
        logger.info(
            f"Running text detection AI model for {len(step_text_detector_images)} images..."
        )
        try:
            ctm.model2annotations_gui(
                profile.general,
                profile.text_detector,
                text_detector_model_path,
                step_text_detector_images,
                cache_dir,
                no_text_detection=False,
                partial_progress_data=partial(
                    imf.ProgressData,
                    len(step_text_detector_images),
                    target_outputs,
                    imf.Step.text_detection,
                ),
                progress_callback=progress_callback,
                abort_flag=abort_flag,
            )
        except wt.Abort:
            # Send the progress callback a signal that the process was aborted,
            # and then raise the exception again to stop the processing.
            check_abortion()

        # Update the outputs of the image objects.
        for image_obj in step_text_detector_images:
            update_output(image_obj, imf.Output.input, ".png")
            update_output(image_obj, imf.Output.ai_mask, "_mask.png")
            update_output(image_obj, imf.Output.raw_json, "#raw.json")

    # ============================================== Preprocessing ==============================================

    check_abortion()

    logger.info(f"Running preprocessing for {len(image_objects)} images...")

    progress_callback.emit(
        imf.ProgressData(
            len(image_objects),
            target_outputs,
            imf.Step.preprocessor,
            imf.ProgressType.begin_step,
        )
    )

    ocr_analytics: list[st.OCRAnalytic] = []
    # Find all the json files associated with the images.
    for image_obj in image_objects:
        check_abortion()
        json_file_path = cache_dir / f"{image_obj.uuid}_{image_obj.path.stem}#raw.json"
        ocr_analytic = pp.prep_json_file(
            json_file_path,
            preprocessor_conf=profile.preprocessor,
            cache_masks=False,
            mocr=ocr_model if profile.preprocessor.ocr_enabled else None,
            cache_masks_ocr=True,
            performing_ocr=True,
        )

        if ocr_analytic is not None:
            ocr_analytics.append(ocr_analytic)

        progress_callback.emit(
            imf.ProgressData(
                len(image_objects),
                target_outputs,
                imf.Step.preprocessor,
                imf.ProgressType.incremental,
            )
        )

    # Update only the raw boxes, the rest are tainted by the forced profile changes.
    for image_obj in image_objects:
        update_output(image_obj, imf.Output.initial_boxes, "_boxes.png")

    logger.info(f"Finished processing {len(image_objects)} images.")

    # ============================================== Final Output ==============================================

    check_abortion()

    # Output the OCRed text from the analytics.
    # Format of the analytics:
    # number of boxes | sizes of all boxes | sizes of boxes that were OCRed | path to image, text, box coordinates
    # We do not need to show the first three columns, so we simplify the data structure.
    path_texts_coords: list[tuple[Path, str, st.Box]] = list(
        itertools.chain.from_iterable(a.removed_box_data for a in ocr_analytics)
    )
    if path_texts_coords:
        paths, texts, boxes = zip(*path_texts_coords)
        paths = hp.trim_prefix_from_paths(paths)
        path_texts_coords = list(zip(paths, texts, boxes))
        # Sort by path.
        path_texts_coords = natsorted(path_texts_coords, key=lambda x: x[0])

    buffer = StringIO()
    if csv_output:
        buffer.write("filename,startx,starty,endx,endy,text\n")

        for path, bubble, box in path_texts_coords:
            path = str(path)
            # Escape commas where necessary.
            if "," in path:
                path = f'"{path}"'

            if "," in bubble:
                bubble = f'"{bubble}"'

            if "\n" in bubble:
                logger.warning(f"Detected newline in bubble: {path} {bubble} {box}")
                bubble = bubble.replace("\n", "\\n")

            buffer.write(f"{path},{box},{bubble}\n")
        text_out = buffer.getvalue()
    else:
        # Place the file path on it's own line, and only if it's different from the previous one.
        current_path = ""
        for path, bubble, _ in path_texts_coords:
            if path != current_path:
                buffer.write(f"\n\n{path}: ")
                current_path = path
            buffer.write(f"\n{bubble}")
            if "\n" in bubble:
                logger.warning(f"Detected newline in bubble: {path} {bubble}")

        text_out = buffer.getvalue()

    text_out = text_out.strip("\n \t")

    if output_file is not None:
        try:
            output_file.parent.mkdir(parents=True, exist_ok=True)
            output_file.write_text(text_out, encoding="utf-8")
            buffer.write(f"\n\nSaved detected text to {output_file}")
        except OSError as e:
            buffer.write(f"\n\nFailed to write detected text to {output_file}")
            logger.exception(e)

    progress_callback.emit(
        imf.ProgressData(
            0,
            [],
            imf.Step.preprocessor,
            imf.ProgressType.outputOCR,
            buffer.getvalue(),
        )
    )

    progress_callback.emit(
        imf.ProgressData(
            0,
            [],
            imf.Step.output,
            imf.ProgressType.end,
        )
    )



import shutil
from functools import partial
from pathlib import Path
from typing import Callable
from typing import Sequence

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Slot, Signal
from loguru import logger

import pcleaner.config as cfg
import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.processing as prc
import pcleaner.gui.structures as st
import pcleaner.gui.worker_thread as wt
from pcleaner.gui.CustomQ.CBadgeButton import BadgeButton
from pcleaner.gui.ui_generated_files.ui_ImageDetails import Ui_ImageDetails


THUMBNAIL_SIZE = 180, 180
PUSHBUTTON_THUMBNAIL_MARGIN = 8
PUSHBUTTON_THUMBNAIL_SEPARATION = 8
SIDEBAR_COLUMNS = 2
BADGE_SIZE = 24


class ImageDetailsWidget(Qw.QWidget, Ui_ImageDetails):
    """
    A widget that shows the details of an image object.
    """

    image_obj: imf.ImageFile  # The image object to show.
    current_image_path: Path | None  # The path of the image currently shown.

    button_map: dict[BadgeButton, imf.Output]

    config: cfg.Config
    shared_ocr_model: st.Shared[st.OCRModel]  # Must be handed over by the file table.
    thread_queue: Qc.QThreadPool
    progress_callback: Callable[[imf.ProgressData], None]
    abort_signal: Signal

    menu: Qw.QMenu  # The overflow menu for the export and ocr buttons.
    export_action: Qg.QAction
    ocr_action: Qg.QAction

    # Used so that the view will zoom to fit on the first load only.
    # This action needs to be delayed since the image dimensions are not known until the image is loaded,
    # since the size may change depending on the scale value.
    first_load: bool

    def __init__(
        self,
        parent=None,
        image_obj: imf.ImageFile = None,
        config: cfg.Config = None,
        shared_ocr_model: st.Shared[st.OCRModel] = None,
        thread_queue: Qc.QThreadPool = None,
        progress_callback: Callable[[imf.ProgressData], None] = None,
        profile_changed_signal: Signal = None,
        abort_signal: Signal = None,
    ):
        """
        Init the widget.

        :param image_obj: The image object to show.
        :param config: The config object.
        :param shared_ocr_model: The shared OCR model.
        :param thread_queue: The thread queue to use for the workers.
        :param progress_callback: The callback to use for the progress signals.
        :param profile_changed_signal: The signal that is emitted when the profile changes.
        :param abort_signal: The signal that is emitted when the worker should abort.
        """
        logger.info(f"Opening details tab for {image_obj.path}")
        Qw.QWidget.__init__(self, parent)
        self.setupUi(self)

        self.image_obj = image_obj
        self.config = config
        self.shared_ocr_model = shared_ocr_model
        self.button_map = self.create_sidebar_buttons()

        self.first_load = True
        self.thread_queue = thread_queue
        self.progress_callback = progress_callback
        self.abort_signal = abort_signal

        # Clear sample text from labels that only update on user interaction.
        self.label_size.setText("")
        self.label_position.setText("0, 0")
        self.label_step.setText("")

        self.init_sidebar()
        self.load_all_image_thumbnails()
        self.pushButton_refresh.clicked.connect(self.regenerate_current_output)
        self.init_menu()
        profile_changed_signal.connect(self.start_profile_checker)
        # Click on the first button to show the input image.
        input_button = list(self.button_map.keys())[0]
        input_button.click()

    def init_menu(self) -> None:
        """
        Initialize the overflow menu housing the export and ocr button for the image details widget.
        """
        self.menu = Qw.QMenu(self)
        self.export_action = Qg.QAction(Qg.QIcon.fromTheme("document-save"), "Export Image", self)
        self.export_action.triggered.connect(self.export_image)
        self.menu.addAction(self.export_action)

        self.ocr_action = Qg.QAction(Qg.QIcon.fromTheme("document-scan"), "OCR Image", self)
        self.ocr_action.triggered.connect(self.start_ocr_worker)
        self.menu.addAction(self.ocr_action)

        self.pushButton_menu.setMenu(self.menu)

    def create_sidebar_buttons(self) -> dict[BadgeButton, imf.Output]:
        """
        Parse the image object's outputs to create a list of buttons in the side panel.

        Returns: A map of buttons to outputs.
        """

        current_button_layout: Qw.QGridLayout
        current_button_index: int = 0
        last_step_name: str | None = None

        def add_button(title: str | None) -> BadgeButton | Qw.QVBoxLayout:
            # Insert a new button into the current step's layout.
            button = BadgeButton()
            button.setCheckable(True)
            button.set_badge_size(BADGE_SIZE, PUSHBUTTON_THUMBNAIL_MARGIN)
            button.clicked.connect(partial(self.uncheck_all_other_buttons, button))
            button_widget = button

            # Create a title as a separate label widget and stack them vertically.
            if title is not None:
                layout = Qw.QVBoxLayout()
                layout.setContentsMargins(0, 0, 0, 0)
                layout.setSpacing(2)  # The small spacing between the label and the image.
                label = Qw.QLabel(title)
                layout.addWidget(label)
                layout.addWidget(button)
                # Wrap the layout in a widget so it can be added to the grid layout.
                button_widget = Qw.QWidget()
                button_widget.setLayout(layout)

            # Add it to the current button grid layout, wrapping to the next row if necessary.
            nonlocal current_button_layout, current_button_index
            current_button_layout.addWidget(
                button_widget,
                current_button_index // SIDEBAR_COLUMNS,
                current_button_index % SIDEBAR_COLUMNS,
            )
            current_button_index += 1
            return button

        def add_step_label(title: str) -> None:
            # Add a step name label with a spacer above it if needed.
            nonlocal last_step_name
            if last_step_name is not None:
                spacer = Qw.QSpacerItem(0, 16, Qw.QSizePolicy.Minimum, Qw.QSizePolicy.Minimum)
                self.sidebar_layout.addItem(spacer)

            label = Qw.QLabel(title)
            font = label.font()
            font.setBold(True)
            label.setFont(font)
            self.sidebar_layout.addWidget(label)
            # Add a fresh grid layout for the buttons.
            nonlocal current_button_layout, current_button_index
            current_button_layout = Qw.QGridLayout()
            current_button_layout.setContentsMargins(0, 0, 0, 0)
            current_button_layout.setSpacing(PUSHBUTTON_THUMBNAIL_SEPARATION)
            current_button_index = 0
            self.sidebar_layout.addLayout(current_button_layout)

        buttons = {}
        for output, proc_output in self.image_obj.outputs.items():
            # Add a step name label if it has changed and is not None.
            # Outputs with a none step name are not shown.
            if proc_output.step_name is None:
                continue
            elif proc_output.step_name != last_step_name:
                add_step_label(proc_output.step_name)
                last_step_name = proc_output.step_name

            button_title = proc_output.output_name
            new_button = add_button(button_title)
            buttons[new_button] = output

        return buttons

    def changeEvent(self, event: Qc.QEvent) -> None:
        """
        Due to the buttons using a fixed stylesheet to apply the colored border as a background,
        they do not react to changes in the palette. Therefore we need to nuke all buttons and
        re-create them when the palette changes.
        :param event:
        :return:
        """
        Qw.QWidget.changeEvent(self, event)
        if event.type() == Qc.QEvent.PaletteChange:
            logger.warning("Palette changed. Re-creating buttons.")

            # Remember what output was last selected.
            current_output = self.button_map.get(self.current_button(), None)

            # Empty out the sidebar layout.
            self._clear_all_from_layout(self.sidebar_layout)
            # Re-create the buttons.
            self.button_map = self.create_sidebar_buttons()
            self.init_sidebar()
            self.load_all_image_thumbnails()

            # Select the button that was selected before.
            for button, output in self.button_map.items():
                if output == current_output:
                    button.click()
                    break

            self.update()

    def _clear_all_from_layout(self, layout: Qw.QLayout) -> None:
        for i in reversed(
            range(layout.count())
        ):  # Loop backwards to avoid disrupting the remaining indices
            item = layout.itemAt(i)
            if item.widget() is not None:
                widget = item.widget()
                widget.deleteLater()  # Mark the widget for deletion
            elif item.layout() is not None:
                # Recursively clear this sub-layout
                self._clear_all_from_layout(item.layout())
            layout.removeItem(item)  # Remove the item from layout

    def init_sidebar(self) -> None:
        """
        Set up the buttons in the sidebar.
        """
        self.label_file_name.setText(self.image_obj.path.name)
        self.label_file_name.setToolTip(str(self.image_obj.path))
        self.label_file_name.setElideMode(Qc.Qt.ElideLeft)

        self.pushButton_zoom_in.clicked.connect(self.image_viewer.zoom_in)
        self.pushButton_zoom_out.clicked.connect(self.image_viewer.zoom_out)
        self.pushButton_zoom_reset.clicked.connect(self.image_viewer.zoom_reset)
        self.pushButton_zoom_fit.clicked.connect(self.image_viewer.zoom_fit)

        # Figure out the optimal button size.
        # It should fit within the thumbnail size while maintaining the aspect ratio.
        # The button size is the minimum of the thumbnail size and the image size.
        image_width, image_height = self.image_obj.size
        ratio = image_width / image_height
        if image_width > image_height:
            thumbnail_width = THUMBNAIL_SIZE[0]
            thumbnail_height = int(thumbnail_width / ratio)
        else:
            thumbnail_height = THUMBNAIL_SIZE[1]
            thumbnail_width = int(thumbnail_height * ratio)

        # Adjust the width of the scroll area to fit the buttons.
        scrollbar_width = Qw.QApplication.style().pixelMetric(Qw.QStyle.PM_ScrollBarExtent)
        margins_left, _, margins_right, _ = self.sidebar_layout.getContentsMargins()
        self.scrollArea.setFixedWidth(
            thumbnail_width * SIDEBAR_COLUMNS
            + PUSHBUTTON_THUMBNAIL_MARGIN * SIDEBAR_COLUMNS
            + PUSHBUTTON_THUMBNAIL_SEPARATION * (SIDEBAR_COLUMNS - 1)
            + scrollbar_width
            + margins_left
            + margins_right
            + 2,  # Arbitrary extra margin to make it look "right".
        )
        logger.debug(f"Setting scroll area width to {self.scrollArea.width()}")

        # Resize all the buttons.
        # self.scrollArea.setFixedWidth()
        for button in self.button_map:
            button.setFixedSize(
                thumbnail_width + PUSHBUTTON_THUMBNAIL_MARGIN,
                thumbnail_height + PUSHBUTTON_THUMBNAIL_MARGIN,
            )
            button.setIconSize(Qc.QSize(thumbnail_width, thumbnail_height))
            button.clicked.connect(partial(self.switch_to_image, button))

    def load_all_image_thumbnails(self) -> None:
        """
        Load all the images into the buttons.
        """
        for button, output in self.button_map.items():
            proc_output = self.image_obj.outputs[output]
            if proc_output.path is not None:
                try:
                    button.setIcon(Qg.QIcon(str(proc_output.path)))
                    button.setText("")
                except OSError as e:
                    logger.error(f"Failed to load image {proc_output.path}: {e}")
            else:
                button.setText("Generate Me")
        # Update the badges on the buttons.
        self.start_profile_checker()

    def switch_to_image(self, button: BadgeButton) -> None:
        """
        Show the image in the button in the image view.

        :param button: The button that was clicked.
        """
        output = self.button_map[button]
        proc_output: imf.ProcessOutput = self.image_obj.outputs[output]
        self.label_step.setText(proc_output.description)
        self.current_image_path = proc_output.path
        if not proc_output.has_path():
            # Clear whatever image is currently shown.
            self.image_viewer.set_image(None)
            self.label_position.setText("0, 0")
            self.widget_footer_info.hide()
            self.stackedWidget.setCurrentWidget(self.page_no_image)
            self.export_action.setEnabled(False)
            self.pushButton_refresh.setEnabled(False)
            self.start_output_worker(output)
        else:
            try:
                # Draw a background using the accent color to provide contrast.
                self.image_viewer.setBackgroundBrush(
                    Qg.QBrush(self.palette().color(Qg.QPalette.Highlight))
                )
                self.image_viewer.set_image(proc_output.path)
                self.label_size.setText(
                    f"{self.image_viewer.image_dimensions[0]} × {self.image_viewer.image_dimensions[1]}"
                )
                self.image_viewer.mouseMoved.connect(self.update_position_label)
                self.widget_footer_info.show()
                self.stackedWidget.setCurrentWidget(self.page_viewer)
                self.export_action.setEnabled(True)
                self.pushButton_refresh.setEnabled(True)
                if self.first_load:
                    self.first_load = False
                    # Call the zoom fit function after the event loop has finished.
                    # This is necessary because the viewport dimensions are not known until the image is loaded.
                    Qc.QTimer.singleShot(0, self.image_viewer.zoom_fit)
            except OSError as e:
                logger.error(f"Image at {proc_output.path} does not exist. {e}")
                gu.show_warning(
                    self, "Image not found.", f"Image at {proc_output.path} does not exist: {e}"
                )
                return

    @Slot(int, int)
    def update_position_label(self, x: int, y: int) -> None:
        """
        Update the position label with the current mouse position.

        :param x: The x coordinate of the mouse.
        :param y: The y coordinate of the mouse.
        """
        self.label_position.setText(f"{x}, {y}")
        self.label_position.setMinimumWidth(self.label_position.width())

    def export_image(self) -> None:
        """
        If the image exists, open a save dialog to copy it to a new location.
        """
        if self.current_image_path is None:
            return
        if not self.current_image_path.is_file():
            return

        save_path, _ = Qw.QFileDialog.getSaveFileName(self, "Export Image", filter="PNG (*.png)")
        if save_path:
            save_path = Path(save_path).with_suffix(".png")
            try:
                shutil.copy(self.current_image_path, save_path)
                logger.info(f"Exported image to {save_path}")
            except OSError as e:
                logger.error(f"Failed to export image to {save_path}: {e}")
                gu.show_warning(self, "Export failed", f"Failed to export image:\n\n{e}")

    def current_button(self) -> BadgeButton | None:
        """
        Get the currently selected button.

        Returns: The button that is currently selected, or None if no button is selected.
        """
        for button in self.button_map:
            if button.isChecked():
                return button
        return None

    def reload_current_image(self) -> None:
        """
        Reload the current image.
        The current image is the one belonging to the currently selected pushbutton.
        """
        if button := self.current_button():
            self.switch_to_image(button)

    @Slot()
    def uncheck_all_other_buttons(self, checked_button: BadgeButton) -> None:
        """
        Uncheck all buttons except the one that was checked.

        :param checked_button: The button that was checked.
        """
        for button in self.button_map:
            if button is not checked_button:
                button.setChecked(False)

    def regenerate_current_output(self) -> None:
        """
        Delete the current output and regenerate it.
        """
        if button := self.current_button():
            self.image_obj.outputs[self.button_map[button]].reset()
            self.switch_to_image(button)

    def clear_change_flairs(self) -> None:
        """
        Clear the change flairs from all buttons.
        """
        for button in self.button_map:
            button.hide_badge()

    @staticmethod
    def set_change_flair(button: BadgeButton) -> None:
        """
        Set the change flair on the given button.

        :param button: The button to set the flair on.
        """
        button.show_badge()

    # ========================================== Worker Functions ==========================================

    def start_output_worker(self, output: imf.Output) -> None:
        """
        Start the worker thread for the given output.

        :param output: The output to generate.
        """
        worker = wt.Worker(self.generate_output, output, abort_signal=self.abort_signal)
        worker.signals.progress.connect(self.progress_callback)
        worker.signals.result.connect(self.output_worker_result)
        worker.signals.error.connect(self.output_worker_error)
        worker.signals.aborted.connect(self.output_worker_aborted)
        self.thread_queue.start(worker)

    def generate_output(
        self, output: imf.Output, progress_callback: imf.ProgressSignal, abort_flag: wt.SharableFlag
    ) -> None:
        """
        Generate the given output, if there doesn't yet exist a valid output for it.

        :param output: The output to generate.
        :param progress_callback: The callback given by the worker thread wrapper.
        :param abort_flag: The flag that is set when the worker should abort.
        """
        # Check if the output is already valid.
        proc_output: imf.ProcessOutput = self.image_obj.outputs[output]
        if proc_output.is_unchanged(self.config.current_profile):
            logger.debug(f"Output {output} is unchanged. Worker will not run.")
            return

        # Start the processor.
        prc.generate_output(
            image_objects=[self.image_obj],
            target_outputs=[output],
            output_dir=None,
            config=self.config,
            ocr_model=self.shared_ocr_model.get(),
            progress_callback=progress_callback,
            abort_flag=abort_flag,
        )

    def output_worker_result(self) -> None:
        self.load_all_image_thumbnails()
        self.reload_current_image()
        logger.info("Output worker finished.")

    @Slot(wt.WorkerError)
    def output_worker_error(self, error: wt.WorkerError) -> None:
        logger.error("Output worker encountered an error.")
        gu.show_warning(self, "Output failed", f"Output generation failed:\n\n{error}")

    def output_worker_aborted(self) -> None:
        self.load_all_image_thumbnails()
        self.reload_current_image()
        self.thread_queue.clear()
        self.pushButton_refresh.setEnabled(True)
        logger.warning("Output worker aborted.")

    def start_profile_checker(self) -> None:
        """
        Start the profile checker thread.
        The process isn't very intensive, usually taking between 1 and 10 ms, but when run on the
        queue it prevents race conditions with the output worker.
        """
        worker = wt.Worker(self.check_profile_changes, no_progress_callback=True)
        worker.signals.error.connect(self.profile_checker_error)
        worker.signals.result.connect(self.profile_checker_result)
        self.thread_queue.start(worker)

    def check_profile_changes(self) -> Sequence[BadgeButton]:
        """
        Check if the current profile has changed for the outputs that exist.
        The regular ordered outputs progressively get more sensitive, so we can
        figure out when the first change happens and know that everything after
        that has changed as well, and vise versa, if something near the end is
        unchanged, we know that everything before it is unchanged as well.

        This doesn't work for the individually sensitive outputs, so we just check
        each of them.
        """

        logger.info("Profile changed. Checking for changes.")

        changed_buttons: list[BadgeButton] = []

        # Create lists of the buttons that have an output to check.
        independent_buttons: list[tuple[BadgeButton, imf.ProcessOutput]] = []
        ordered_buttons: list[tuple[BadgeButton, imf.ProcessOutput]] = []

        for button, output in self.button_map.items():
            # Check if the output is individually sensitive or has no path.
            proc_output: imf.ProcessOutput = self.image_obj.outputs[output]

            if proc_output.path is None:
                continue

            if output in imf.OUTPUTS_WITH_INDEPENDENT_PROFILE_SENSITIVITY:
                independent_buttons.append((button, proc_output))
            else:
                ordered_buttons.append((button, proc_output))

        # Handle the independent buttons first.
        for button, proc_output in independent_buttons:
            if not proc_output.is_unchanged(self.config.current_profile):
                changed_buttons.append(button)

        # Sanity check, do we even have any ordered buttons?
        if len(ordered_buttons) == 0:
            return changed_buttons

        # Perform a binary search on the ordered buttons, to find the first changed output.
        # But first check the bounds, first and last.
        # Cache the check results in a tri-state list. We can't use a typical cache because
        # the profile will change between runs.
        ordered_buttons_changed: list[bool | None] = [None] * len(ordered_buttons)

        # Keep track of the number of cache misses for the check function.
        change_check_calls = 0
        change_check_cache_misses = 0

        def check_changed(index: int) -> bool:
            nonlocal ordered_buttons_changed, ordered_buttons, self
            nonlocal change_check_cache_misses, change_check_calls
            change_check_calls += 1
            # logger.debug(
            #     f"Checking button {index}, with output {self.button_map[ordered_buttons[index][0]].name}"
            # )
            if (changed := ordered_buttons_changed[index]) is not None:
                return changed
            changed = ordered_buttons[index][1].is_changed(self.config.current_profile)
            change_check_cache_misses += 1
            ordered_buttons_changed[index] = changed
            return changed

        # Check the first and last buttons.
        if not check_changed(-1):
            # logger.debug(
            #     f"Determined that none of the ordered buttons changed in {change_check_calls} calls"
            #     f" ({1-change_check_cache_misses/change_check_calls:.0%} cache hits)."
            # )
            # Nothing is changed.
            return changed_buttons
        elif check_changed(0):
            # Everything is changed.
            ordered_buttons_changed = [True] * len(ordered_buttons)
        else:
            # Now we do the binary search, knowing that a change
            # happened somewhere between the first and last.
            low, high = 0, len(ordered_buttons) - 1
            current: int = (low + high) // 2

            # The logic of the search is as follows:
            # If the current button has changed but the previous one hasn't, then
            # current points to the first changed button.
            # Otherwise, if the current button has not changed, move low to current + 1.
            # If the current button has changed, move high to current - 1.
            while low <= high:
                # Already handled the boundary cases, so we know that the current button's
                # previous button exists and is the last unchanged one.
                if check_changed(current) and not check_changed(current - 1):
                    break
                elif not check_changed(current):
                    low = current + 1
                else:
                    high = current - 1
                current = (low + high) // 2

            # At this point, current points to the first changed button.
            # Everything from this point onwards is also changed,
            # and everything before it is unchanged.
            for idx in range(len(ordered_buttons_changed)):
                ordered_buttons_changed[idx] = idx >= current

            # logger.debug(
            #     f"Determined that none of the ordered buttons changed in {change_check_calls} calls"
            #     f" ({1-change_check_cache_misses/change_check_calls:.0%} cache hits)."
            # )

        # Update the flair for the buttons that have changed.
        for idx, (button, _) in enumerate(ordered_buttons):
            if ordered_buttons_changed[idx]:
                changed_buttons.append(button)

        return changed_buttons

    def profile_checker_result(self, changed_buttons: Sequence[BadgeButton]) -> None:
        """
        Update the flair on the buttons that have changed.

        :param changed_buttons: The buttons that have changed.
        """
        self.clear_change_flairs()
        # logger.debug(
        #     f"Setting change badge for "
        #     f"{len(changed_buttons)} buttons: {','.join(self.button_map[output].name for output in changed_buttons)}"
        # )
        for button in changed_buttons:
            self.set_change_flair(button)

    def profile_checker_error(self, error: wt.WorkerError) -> None:
        logger.error("Profile checker encountered an error.")
        gu.show_warning(self, "Profile check failed", f"Profile change check failed:\n\n{error}")

    def start_ocr_worker(self) -> None:
        """
        Start a worker to perform the OCR process on this image.
        """
        self.ocr_action.setEnabled(False)
        worker = wt.Worker(self.generate_ocr, abort_signal=self.abort_signal)
        worker.signals.progress.connect(self.progress_callback)
        worker.signals.finished.connect(self.ocr_worker_finished)
        worker.signals.result.connect(self.output_worker_result)
        worker.signals.error.connect(self.output_worker_error)
        self.thread_queue.start(worker)

    def generate_ocr(
        self, progress_callback: imf.ProgressSignal, abort_flag: wt.SharableFlag
    ) -> None:
        """
        Run OCR for the image.

        :param progress_callback: The callback given by the worker thread wrapper.
        :param abort_flag: The flag that is set when the worker should abort.
        """
        prc.perform_ocr(
            image_objects=[self.image_obj],
            output_file=None,
            csv_output=False,
            config=self.config,
            ocr_model=self.shared_ocr_model.get(),
            progress_callback=progress_callback,
            abort_flag=abort_flag,
        )

    def ocr_worker_finished(self) -> None:
        self.ocr_action.setEnabled(True)
        logger.info("OCR worker finished.")



from pathlib import Path
from typing import Callable

import PySide6.QtCore as Qc
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Slot
from loguru import logger

import pcleaner.config as cfg
import pcleaner.gui.image_details_driver as idd
import pcleaner.gui.image_file as imf
import pcleaner.gui.structures as st


# noinspection PyPep8Naming
class ImageTab(Qw.QTabWidget):
    """
    Manage the static tab for the file table and the dynamic tabs showing the image
    previews/outputs.
    """

    # Currently open dynamic tabs: path -> (image structure, tab's widget)
    open_images: dict[Path, tuple[imf.ImageFile, Qw.QWidget]]

    def __init__(self, parent=None) -> None:
        Qw.QTabWidget.__init__(self, parent)
        self.setAcceptDrops(True)

        self.open_images = {}

        self.tabCloseRequested.connect(self.tab_close)

    @Slot(imf.ImageFile)
    def open_image(
        self,
        image_obj: imf.ImageFile,
        config: cfg.Config,
        shared_ocr_model: st.Shared[st.OCRModel],
        thread_queue: Qc.QThreadPool,
        progress_callback: Callable[[imf.ProgressData], None],
        profile_changed_signal: Qc.Signal,
        abort_signal: Qc.Signal,
    ):
        """
        Check if the image is already open, in which case show it.
        Otherwise create a new tab.

        :param image_obj: Image object to open.
        :param config: The config object.
        :param shared_ocr_model: The shared OCR model.
        :param thread_queue: The thread queue for processing steps.
        :param progress_callback: The callback to call when a step is done.
        :param profile_changed_signal: The signal that is broadcast when the profile is changed.
        :param abort_signal: The signal that is broadcast when the user aborts the process.
        """
        if image_obj.path in self.open_images:
            self.setCurrentWidget(self.open_images[image_obj.path][1])
            return

        # Create the tab.
        tab = idd.ImageDetailsWidget(
            parent=self,
            image_obj=image_obj,
            config=config,
            shared_ocr_model=shared_ocr_model,
            thread_queue=thread_queue,
            progress_callback=progress_callback,
            profile_changed_signal=profile_changed_signal,
            abort_signal=abort_signal,
        )
        self.addTab(tab, image_obj.path.name)
        self.open_images[image_obj.path] = (image_obj, tab)
        self.setCurrentWidget(tab)

    @Slot(int)
    def tab_close(self, index: int) -> None:
        """
        Close the image details tab.

        :param index: The index of the tab to close.
        """
        # Make sure the index is not the primary tab.
        if index == 0:
            logger.warning("Attempted to close the primary tab.")
            return
        logger.debug(f"Closing tab at index {index}.")
        # The tab we're closing must be an image details tab.
        # noinspection PyTypeChecker
        widget_to_close: idd.ImageDetailsWidget = self.widget(index)
        path = widget_to_close.image_obj.path
        self.open_images.pop(path)
        self.removeTab(index)

    def clear_files(self) -> None:
        """
        Clear all files from the table, removing all tabs except the primary tab.
        """
        self.open_images.clear()
        for i in range(self.count() - 1, 0, -1):
            self.removeTab(i)



import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Signal


# noinspection PyPep8Naming
class CDropFrame(Qw.QFrame):
    """
    Extends the functionality with custom helpers
    And includes a secondary array for data linked to each item
    """

    drop_signal = Signal(Qg.QDropEvent)

    def __init__(self, parent=None) -> None:
        Qw.QFrame.__init__(self, parent)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event: Qg.QDragEnterEvent) -> None:
        if event.mimeData().hasUrls():
            event.acceptProposedAction()

    def dropEvent(self, event: Qg.QDropEvent) -> None:
        self.drop_signal.emit(event)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ImageDetails.ui'
##
## Created by: Qt User Interface Compiler version 6.5.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGraphicsView, QHBoxLayout,
    QLabel, QLayout, QPushButton, QScrollArea,
    QSizePolicy, QStackedWidget, QVBoxLayout, QWidget)

from pcleaner.gui.CustomQ.CElidedLabel import CElidedLabel
from pcleaner.gui.image_viewer import ImageViewer

class Ui_ImageDetails(object):
    def setupUi(self, ImageDetails):
        if not ImageDetails.objectName():
            ImageDetails.setObjectName(u"ImageDetails")
        ImageDetails.resize(900, 800)
        self.horizontalLayout_3 = QHBoxLayout(ImageDetails)
        self.horizontalLayout_3.setSpacing(0)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.scrollArea = QScrollArea(ImageDetails)
        self.scrollArea.setObjectName(u"scrollArea")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy)
        self.scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 138, 943))
        self.horizontalLayout_2 = QHBoxLayout(self.scrollAreaWidgetContents)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setSizeConstraint(QLayout.SetMinimumSize)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.widget = QWidget(self.scrollAreaWidgetContents)
        self.widget.setObjectName(u"widget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy1)
        self.sidebar_layout = QVBoxLayout(self.widget)
        self.sidebar_layout.setObjectName(u"sidebar_layout")
        self.sidebar_layout.setContentsMargins(4, 4, 4, 4)

        self.horizontalLayout_2.addWidget(self.widget)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.horizontalLayout_3.addWidget(self.scrollArea)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(6, 4, 4, 4)
        self.label_file_name = CElidedLabel(ImageDetails)
        self.label_file_name.setObjectName(u"label_file_name")
        sizePolicy2 = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.label_file_name.sizePolicy().hasHeightForWidth())
        self.label_file_name.setSizePolicy(sizePolicy2)
        self.label_file_name.setFrameShape(QFrame.NoFrame)
        self.label_file_name.setFrameShadow(QFrame.Raised)

        self.horizontalLayout.addWidget(self.label_file_name)

        self.pushButton_zoom_in = QPushButton(ImageDetails)
        self.pushButton_zoom_in.setObjectName(u"pushButton_zoom_in")
        icon = QIcon()
        iconThemeName = u"zoom-in"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_zoom_in.setIcon(icon)
        self.pushButton_zoom_in.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_in)

        self.pushButton_zoom_out = QPushButton(ImageDetails)
        self.pushButton_zoom_out.setObjectName(u"pushButton_zoom_out")
        icon1 = QIcon()
        iconThemeName = u"zoom-out"
        if QIcon.hasThemeIcon(iconThemeName):
            icon1 = QIcon.fromTheme(iconThemeName)
        else:
            icon1.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_zoom_out.setIcon(icon1)
        self.pushButton_zoom_out.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_out)

        self.pushButton_zoom_reset = QPushButton(ImageDetails)
        self.pushButton_zoom_reset.setObjectName(u"pushButton_zoom_reset")
        icon2 = QIcon()
        iconThemeName = u"zoom-original"
        if QIcon.hasThemeIcon(iconThemeName):
            icon2 = QIcon.fromTheme(iconThemeName)
        else:
            icon2.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_zoom_reset.setIcon(icon2)
        self.pushButton_zoom_reset.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_reset)

        self.pushButton_zoom_fit = QPushButton(ImageDetails)
        self.pushButton_zoom_fit.setObjectName(u"pushButton_zoom_fit")
        icon3 = QIcon()
        iconThemeName = u"zoom-fit-best"
        if QIcon.hasThemeIcon(iconThemeName):
            icon3 = QIcon.fromTheme(iconThemeName)
        else:
            icon3.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_zoom_fit.setIcon(icon3)
        self.pushButton_zoom_fit.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_fit)

        self.pushButton_refresh = QPushButton(ImageDetails)
        self.pushButton_refresh.setObjectName(u"pushButton_refresh")
        self.pushButton_refresh.setEnabled(False)
        icon4 = QIcon()
        iconThemeName = u"view-refresh"
        if QIcon.hasThemeIcon(iconThemeName):
            icon4 = QIcon.fromTheme(iconThemeName)
        else:
            icon4.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_refresh.setIcon(icon4)

        self.horizontalLayout.addWidget(self.pushButton_refresh)

        self.pushButton_menu = QPushButton(ImageDetails)
        self.pushButton_menu.setObjectName(u"pushButton_menu")
        sizePolicy3 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.pushButton_menu.sizePolicy().hasHeightForWidth())
        self.pushButton_menu.setSizePolicy(sizePolicy3)
        self.pushButton_menu.setMinimumSize(QSize(16, 16))
        self.pushButton_menu.setContextMenuPolicy(Qt.ActionsContextMenu)
        icon5 = QIcon(QIcon.fromTheme(u"overflow-menu"))
        self.pushButton_menu.setIcon(icon5)

        self.horizontalLayout.addWidget(self.pushButton_menu)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.stackedWidget = QStackedWidget(ImageDetails)
        self.stackedWidget.setObjectName(u"stackedWidget")
        self.page_viewer = QWidget()
        self.page_viewer.setObjectName(u"page_viewer")
        self.verticalLayout_3 = QVBoxLayout(self.page_viewer)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.image_viewer = ImageViewer(self.page_viewer)
        self.image_viewer.setObjectName(u"image_viewer")
        sizePolicy4 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.image_viewer.sizePolicy().hasHeightForWidth())
        self.image_viewer.setSizePolicy(sizePolicy4)
        self.image_viewer.setFrameShape(QFrame.NoFrame)
        self.image_viewer.setFrameShadow(QFrame.Plain)
        self.image_viewer.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)

        self.verticalLayout_3.addWidget(self.image_viewer)

        self.stackedWidget.addWidget(self.page_viewer)
        self.page_no_image = QWidget()
        self.page_no_image.setObjectName(u"page_no_image")
        self.verticalLayout_4 = QVBoxLayout(self.page_no_image)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.label_17 = QLabel(self.page_no_image)
        self.label_17.setObjectName(u"label_17")

        self.verticalLayout_4.addWidget(self.label_17, 0, Qt.AlignHCenter)

        self.stackedWidget.addWidget(self.page_no_image)

        self.verticalLayout.addWidget(self.stackedWidget)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setSpacing(12)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(6, 4, 4, 4)
        self.label_step = QLabel(ImageDetails)
        self.label_step.setObjectName(u"label_step")
        sizePolicy2.setHeightForWidth(self.label_step.sizePolicy().hasHeightForWidth())
        self.label_step.setSizePolicy(sizePolicy2)
        self.label_step.setWordWrap(True)

        self.horizontalLayout_4.addWidget(self.label_step)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setSpacing(4)
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.label_position_label = QLabel(ImageDetails)
        self.label_position_label.setObjectName(u"label_position_label")

        self.horizontalLayout_6.addWidget(self.label_position_label)

        self.label_position = QLabel(ImageDetails)
        self.label_position.setObjectName(u"label_position")

        self.horizontalLayout_6.addWidget(self.label_position)


        self.horizontalLayout_4.addLayout(self.horizontalLayout_6)

        self.horizontalLayout_7 = QHBoxLayout()
        self.horizontalLayout_7.setSpacing(4)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.label_size_label = QLabel(ImageDetails)
        self.label_size_label.setObjectName(u"label_size_label")

        self.horizontalLayout_7.addWidget(self.label_size_label)

        self.label_size = QLabel(ImageDetails)
        self.label_size.setObjectName(u"label_size")

        self.horizontalLayout_7.addWidget(self.label_size)


        self.horizontalLayout_4.addLayout(self.horizontalLayout_7)

        self.widget_footer_info = QWidget(ImageDetails)
        self.widget_footer_info.setObjectName(u"widget_footer_info")
        self.horizontalLayout_5 = QHBoxLayout(self.widget_footer_info)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)

        self.horizontalLayout_4.addWidget(self.widget_footer_info)


        self.verticalLayout.addLayout(self.horizontalLayout_4)


        self.horizontalLayout_3.addLayout(self.verticalLayout)


        self.retranslateUi(ImageDetails)

        self.stackedWidget.setCurrentIndex(1)


        QMetaObject.connectSlotsByName(ImageDetails)
    # setupUi

    def retranslateUi(self, ImageDetails):
        ImageDetails.setWindowTitle(QCoreApplication.translate("ImageDetails", u"Form", None))
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_in.setToolTip(QCoreApplication.translate("ImageDetails", u"Zoom in", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_in.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_out.setToolTip(QCoreApplication.translate("ImageDetails", u"Zoom out", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_out.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_reset.setToolTip(QCoreApplication.translate("ImageDetails", u"Reset zoom", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_reset.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_fit.setToolTip(QCoreApplication.translate("ImageDetails", u"Zoom to fit", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_fit.setText("")
        self.pushButton_refresh.setText(QCoreApplication.translate("ImageDetails", u"Refresh", None))
        self.pushButton_menu.setText("")
        self.label_17.setText(QCoreApplication.translate("ImageDetails", u"Generating...", None))
        self.label_step.setText(QCoreApplication.translate("ImageDetails", u"<Current Step>", None))
        self.label_position_label.setText(QCoreApplication.translate("ImageDetails", u"Position:", None))
        self.label_position.setText(QCoreApplication.translate("ImageDetails", u"<x,y>", None))
        self.label_size_label.setText(QCoreApplication.translate("ImageDetails", u"Size:", None))
        self.label_size.setText(QCoreApplication.translate("ImageDetails", u"<w x h>", None))
    # retranslateUi




from typing import Generic, TypeVar, Optional, Protocol

from PIL import Image


T = TypeVar("T")


class Shared(Generic[T]):
    def __init__(self, initial_value: Optional[T] = None) -> None:
        self._container = {"data": initial_value}

    def get(self) -> Optional[T]:
        return self._container["data"]

    def set(self, value: T) -> None:
        self._container["data"] = value

    def is_none(self) -> bool:
        return self._container["data"] is None


class OCRModel(Protocol):
    def __call__(self, image: Image) -> str:
        ...



import PySide6.QtWidgets as Qw
import PySide6.QtGui as Qg
import PySide6.QtCore as Qc


# noinspection PyPep8Naming
class CTableWidget(Qw.QTableWidget):
    """
    Extends the functionality with custom helpers
    """

    finished_drop = Qc.Signal()

    def __init__(self, parent=None) -> None:
        Qw.QTableWidget.__init__(self, parent)

    def clearAll(self) -> None:
        self.clearContents()
        self.setRowCount(0)
        self.itemSelectionChanged.emit()

    def currentText(self, col: int) -> None:
        return self.item(self.currentRow(), col).text()

    def setCurrentText(self, col: int, text: str) -> None:
        return self.item(self.currentRow(), col).setText(text)

    def appendRow(self, *args: str, select_new: bool = False) -> None:
        """
        Adds a new row to the bottom and fills each column with one of the args
        :param args: list(str)
        :param select_new: bool - highlight the new row
        """
        rows = self.rowCount()
        self.setRowCount(rows + 1)
        for i, arg in enumerate(args):
            self.setItem(rows, i, Qw.QTableWidgetItem(arg))
        if select_new:
            self.setCurrentCell(rows, 0)

    def moveRowUp(self, row: int) -> None:
        """
        Inserts a new row above row, then moves all items into it.
        Then deletes old row and selects new row.
        :param row: int
        """
        self.insertRow(row - 1)
        for i in range(self.columnCount()):
            item = self.takeItem(row + 1, i)
            self.setItem(row - 1, i, item)
        self.removeRow(row + 1)
        self.setCurrentCell(row - 1, 0)

    def moveRowDown(self, row: int) -> None:
        """
        Inserts a new row below row, then moves all items into it.
        Then deletes old row and selects new row.
        :param row: int
        """
        self.insertRow(row + 2)
        for i in range(self.columnCount()):
            item = self.takeItem(row, i)
            self.setItem(row + 2, i, item)
        self.removeRow(row)
        self.setCurrentCell(row + 1, 0)

    def columnValues(self, col: int) -> list:
        values = []
        for row in range(self.rowCount()):
            values.append(self.item(row, col).text())
        return values

    def resizeHeightToContents(self) -> None:
        # Add 1 per row to compensate for grid lines. Subtract 1 at the end since only inner grid lines count.
        height = 0
        for i in range(self.rowCount()):
            height += self.rowHeight(i) + 1
        self.setMinimumHeight(height - 1)

    def hasSelected(self) -> bool:
        return self.selectedIndexes() != []

    def dragEnterEvent(self, event: Qg.QDragEnterEvent) -> None:
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event: Qg.QDragMoveEvent) -> None:
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event: Qg.QDropEvent) -> None:
        if event.mimeData().hasUrls():
            for url in event.mimeData().urls():
                self.handleDrop(url.toLocalFile())
            event.accept()
        else:
            event.ignore()
        self.finished_drop.emit()

    def handleDrop(self, path: str) -> None:
        pass



import PySide6.QtWidgets as Qw
import PySide6.QtGui as Qg
import PySide6.QtCore as Qc


class CElidedLabel(Qw.QFrame):
    elisionChanged = Qc.Signal(bool)

    def __init__(self, text, parent=None) -> None:
        super(CElidedLabel, self).__init__(parent)

        self._elided = False
        self._content = text
        self._elideMode = Qc.Qt.ElideRight
        self.setSizePolicy(Qw.QSizePolicy.Expanding, Qw.QSizePolicy.Preferred)
        self.setFrameStyle(Qw.QFrame.NoFrame)

    def text(self) -> None:
        return self._content

    def setText(self, newText) -> None:
        self._content = newText
        self.update()

    def isElided(self) -> bool:
        return self._elided

    def elideMode(self) -> None:
        return self._elideMode

    def setElideMode(self, mode) -> None:
        if mode in [Qc.Qt.ElideLeft, Qc.Qt.ElideMiddle, Qc.Qt.ElideRight, Qc.Qt.ElideNone]:
            self._elideMode = mode
            self.update()
        else:
            raise ValueError(f"Invalid elide mode {mode}")

    def paintEvent(self, event) -> None:
        painter = Qg.QPainter(self)
        fontMetrics = painter.fontMetrics()

        didElide = False
        lineSpacing = fontMetrics.lineSpacing()

        # For vertical centering, calculate the y-position
        y = (self.height() - lineSpacing) // 2 + fontMetrics.ascent()

        textLayout = Qg.QTextLayout(self._content, painter.font())
        textLayout.beginLayout()
        while True:
            line = textLayout.createLine()

            if not line.isValid():
                break

            line.setLineWidth(self.width())

            lastLine = self._content[line.textStart() :]
            elidedLastLine = fontMetrics.elidedText(lastLine, self._elideMode, self.width())
            painter.drawText(Qc.QPoint(0, y), elidedLastLine)
            line = textLayout.createLine()
            didElide = line.isValid()
            break

        textLayout.endLayout()

        if didElide != self._elided:
            self._elided = didElide
            self.elisionChanged.emit(didElide)
